from gwcli.node import UIGroup, UINode

from gwcli.utils import response_message, APIRequest, get_config

from ceph_iscsi_config.client import CHAP, GWClient
import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import human_size, this_host

from rtslib_fb.utils import normalize_wwn, RTSLibError

# this ignores the warning issued when verify=False is used
from requests.packages import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Clients(UIGroup):
    help_intro = '''
                 The host section shows the clients that have been defined
                 across each of the gateways in the configuration.

                 Clients may be created or deleted, but once defined they can
                 not be renamed.

                 e.g.
                 create iqn.1994-05.com.redhat:rh7-client4
                 '''

    def __init__(self, parent):
        UIGroup.__init__(self, 'hosts', parent)

        # lun_map dict is indexed by the rbd name, pointing to a list
        # of clients that have that rbd allocated.
        self.lun_map = {}
        self.client_map = {}

        # record the shortcut
        shortcut = self.shell.prefs['bookmarks'].get('hosts', None)
        if not shortcut or shortcut is not self.path:
            self.shell.prefs['bookmarks']['hosts'] = self.path
            self.shell.prefs.save()
            self.shell.log.debug("Bookmarked %s as %s."
                                 % (self.path, 'hosts'))

        self.config = get_config()

    def load(self, client_info):
        for client_iqn, client_settings in client_info.items():
            Client(self, client_iqn, client_settings)

    def ui_command_create(self, client_iqn):
        """
        Clients may be created using the 'create' sub-command. The initial
        definition will be added to each gateway without any authentication
        set. Once a client is created the admin is automatically placed in the
        context of the new client definition for auth and disk configuration
        operations.

        e.g.
        > create <client_iqn>

        """
        self.logger.debug("CMD: ../hosts/ create {}".format(client_iqn))
        cli_seed = {"luns": {}, "auth": {}}

        # is the IQN usable?
        try:
            client_iqn, iqn_type = normalize_wwn(['iqn'], client_iqn)
        except RTSLibError:
            self.logger.error("IQN name '{}' is not valid for "
                              "iSCSI".format(client_iqn))
            return

        target_iqn = self.parent.name

        # Issue the API call to create the client
        client_api = ('{}://localhost:{}/api/'
                      'client/{}/{}'.format(self.http_mode,
                                            settings.config.api_port,
                                            target_iqn,
                                            client_iqn))

        self.logger.debug("Client CREATE for {}".format(client_iqn))
        api = APIRequest(client_api)
        api.put()

        if api.response.status_code == 200:
            Client(self, client_iqn, cli_seed)
            self.config = get_config()
            self.logger.debug("- Client '{}' added".format(client_iqn))
            self.logger.info('ok')

        else:
            self.logger.error("Failed: {}".format(response_message(api.response,
                                                                   self.logger)))
            return

        # switch the current directory to the new client for auth or disk
        # definitions as part of the users workflow
        return self.ui_command_cd(client_iqn)

    def ui_command_delete(self, client_iqn):
        """
        You may delete a client from the configuration, but you must ensure
        the client has logged out of the iscsi gateways. Attempting to delete a
        client that has an open session will fail the request

        e.g.
        delete <client_iqn>

        """

        self.logger.debug("CMD: ../hosts/ delete {}".format(client_iqn))

        self.logger.debug("Client DELETE for {}".format(client_iqn))

        # is the IQN usable?
        try:
            client_iqn, iqn_type = normalize_wwn(['iqn'], client_iqn)
        except RTSLibError:
            self.logger.error("IQN name '{}' is not valid for "
                              "iSCSI".format(client_iqn))
            return

        target_iqn = self.parent.name

        client_api = ('{}://{}:{}/api/'
                      'client/{}/{}'.format(self.http_mode,
                                            "localhost",
                                            settings.config.api_port,
                                            target_iqn,
                                            client_iqn))
        api = APIRequest(client_api)
        api.delete()

        if api.response.status_code == 200:
            # Delete successful across all gateways
            self.logger.debug("- '{}' removed and configuration "
                              "updated".format(client_iqn))

            client = [client for client in self.children
                      if client.name == client_iqn][0]

            # remove any rbd maps from the lun_map for this client
            rbds_mapped = [lun.rbd_name for lun in client.children]
            for rbd in rbds_mapped:
                self.update_lun_map('remove', rbd, client_iqn)

            self.delete(client)

            self.logger.info('ok')
        else:
            # client delete request failed
            self.logger.error(response_message(api.response, self.logger))

    def update_lun_map(self, action, rbd_path, client_iqn):
        """
        Update the lun_map lookup dict, each element points to a 'set' of
        clients that have the lun mapped
        :param action: add or remove
        :param rbd_path: disk name (str) i.e. <pool>.<rbd_image>
        :param client_iqn: client IQN (str)
        """

        if action == 'add':
            if rbd_path in self.lun_map:
                self.lun_map[rbd_path].add(client_iqn)
            else:
                self.lun_map[rbd_path] = {client_iqn}
        else:
            if rbd_path in self.lun_map:
                try:
                    self.lun_map[rbd_path].remove(client_iqn)
                except KeyError:
                    # client not in set
                    pass
                else:
                    if len(self.lun_map[rbd_path]) == 0:
                        del self.lun_map[rbd_path]
            else:
                # delete requested for an rbd that is not in lun_map?
                # twilight zone moment
                raise ValueError("Clients.update_lun_map : Attempt to delete "
                                 "rbd from lun_map that is not defined")

    def delete(self, child):
        del self.client_map[child.client_iqn]
        self.remove_child(child)

    def ui_command_auth(self, action=None):
        """
        Disable/enable ACL authentication or clear CHAP settings for all clients on the target.

        - disable_acl ... Disable initiator name based ACL authentication.

        - enable_acl .... Enable initiator name based ACL authentication.

        - nochap ........ Remove chap authentication for all clients across all gateways.
                          Initiator name based authentication will then be used.

        e.g.
        auth disable_acl

        """

        if not action:
            self.logger.error("Missing auth argument. Use 'auth nochap|disable_acl|enable_acl'")
            return

        if action not in ['nochap', 'enable_acl', 'disable_acl']:
            self.logger.error("Invalid auth argument. Use 'auth nochap|disable_acl|enable_acl'")
            return

        if action == 'nochap':
            for client in self.children:
                client.set_auth(action, None, None, None)
        else:
            target_iqn = self.parent.name
            api_vars = {'action': action}
            targetauth_api = ('{}://localhost:{}/api/'
                              'targetauth/{}'.format(self.http_mode,
                                                     settings.config.api_port,
                                                     target_iqn))
            api = APIRequest(targetauth_api, data=api_vars)
            api.put()
            if api.response.status_code == 200:
                self.config = get_config()
                self.logger.info('ok')
            else:
                self.logger.error("Failed to {}: "
                                  "{}".format(action, response_message(api.response,
                                                                       self.logger)))
                return

    def summary(self):
        chap_enabled = False
        chap_disabled = False
        target_iqn = self.parent.name

        target_auth = self.parent.auth
        target_auth_enabled = target_auth['username'] and target_auth['password']

        if self.config['targets'][target_iqn]['acl_enabled']:
            auth_stat_str = "ACL_ENABLED"
            status = True
        else:
            auth_stat_str = "ACL_DISABLED"
            status = None if target_auth_enabled else False

        if not target_auth_enabled:
            for client in self.children:
                if not client.auth['username']:
                    chap_disabled = True
                else:
                    chap_enabled = True

                if chap_enabled and chap_disabled:
                    auth_stat_str = "MISCONFIG"
                    status = False
                    break

        return "Auth: {}, Hosts: {}".format(auth_stat_str, len(self.children)),\
            status


class Client(UINode):
    help_intro = '''
                 Client definitions can be managed through two sub-commands;
                 'auth' and 'disk'. These commands allow you to manage the
                 CHAP authentication for the client (1-WAY) and change the
                 rbd images masked to a specific client.

                 LUN masking automatically associates a specific LUN id to the
                 for an rbd image, simplifying the workflow. The lun id's
                 assigned can be seen by running the 'info' command. This will
                 show all the clients details that are stored within the
                 iscsi gateway configuration.
                 '''

    display_attributes = ["client_iqn", "ip_address", "alias", "logged_in",
                          "auth", "group_name", "luns"]

    def __init__(self, parent, client_iqn, client_settings):
        UINode.__init__(self, client_iqn, parent)
        self.client_iqn = client_iqn
        self.parent.client_map[client_iqn] = self
        self.group_name = ''

        self.ip_address = ''
        self.alias = ''

        for k, v in client_settings.items():
            self.__setattr__(k, v)

        # decode the chap password if necessary
        if 'username' in self.auth and 'password' in self.auth:
            self.chap = CHAP(self.auth['username'],
                             self.auth['password'],
                             self.auth['password_encryption_enabled'])
            self.auth['username'] = self.chap.user
            self.auth['password'] = self.chap.password
        else:
            self.auth['username'] = ''
            self.auth['password'] = ''

        # decode the chap_mutual password if necessary
        if 'mutual_username' in self.auth and 'mutual_password' in self.auth:
            self.chap_mutual = CHAP(self.auth['mutual_username'],
                                    self.auth['mutual_password'],
                                    self.auth['mutual_password_encryption_enabled'])
            self.auth['mutual_username'] = self.chap_mutual.user
            self.auth['mutual_password'] = self.chap_mutual.password
        else:
            self.auth['mutual_username'] = ''
            self.auth['mutual_password'] = ''

        self.refresh_luns()

    def drop_luns(self):
        luns = self.children.copy()
        for lun in luns:
            self.remove_lun(lun)

    def refresh_luns(self):
        for rbd_path in self.luns.keys():
            lun_id = self.luns[rbd_path]['lun_id']
            self.parent.update_lun_map('add', rbd_path, self.client_iqn)
            MappedLun(self, rbd_path, lun_id)

    def __str__(self):
        return self.get_info()

    def summary(self):

        all_pools = self.parent.parent.parent.parent.disks.children
        all_disks = []
        for pool in all_pools:
            for disk in pool.children:
                all_disks.append(disk)
        total_bytes = 0

        client_luns = [lun.rbd_name for lun in self.children]

        for disk in all_disks:
            if disk.image_id in client_luns:
                total_bytes += disk.size

        msg = ['LOGGED-IN'] if self.logged_in else []

        auth_text = "Auth: None"
        status = False

        if self.auth.get('mutual_username'):
            auth_text = "Auth: CHAP_MUTUAL"
            status = True
        elif self.auth.get('username'):
            auth_text = "Auth: CHAP"
            status = True

        msg.append(auth_text)

        msg.append("Disks: {}({})".format(len(client_luns),
                                          human_size(total_bytes)))

        return ", ".join(msg), status

    def set_auth(self, username=None, password=None, mutual_username=None, mutual_password=None):

        self.logger.debug("username={}, password={}, mutual_username={}, "
                          "mutual_password={}".format(username, password, mutual_username,
                                                      mutual_password))

        self.logger.debug("CMD: ../hosts/<client_iqn> auth *")

        if not username:
            self.logger.error("To set or reset authentication, specify either "
                              "username=<user> password=<password> "
                              "[mutual_username]=<user> [mutual_password]=<password> "
                              "or nochap")
            return

        if username == 'nochap':
            username = ''
            password = ''
            mutual_username = ''
            mutual_password = ''

        self.logger.debug("auth to be set to username='{}', password='{}', mutual_username='{}', "
                          "mutual_password='{}' for '{}'".format(username, password,
                                                                 mutual_username, mutual_password,
                                                                 self.client_iqn))

        target_iqn = self.parent.parent.name

        api_vars = {
            "username": username,
            "password": password,
            "mutual_username": mutual_username,
            "mutual_password": mutual_password
        }

        clientauth_api = ('{}://localhost:{}/api/'
                          'clientauth/{}/{}'.format(self.http_mode,
                                                    settings.config.api_port,
                                                    target_iqn,
                                                    self.client_iqn))

        api = APIRequest(clientauth_api, data=api_vars)
        api.put()

        if api.response.status_code == 200:
            self.logger.debug("- client credentials updated")
            self.auth['username'] = username
            self.auth['password'] = password
            self.auth['mutual_username'] = mutual_username
            self.auth['mutual_password'] = mutual_password
            self.logger.info('ok')

        else:
            self.logger.error("Failed to update the client's auth: "
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return

    def ui_command_auth(self, username=None, password=None, mutual_username=None,
                        mutual_password=None):
        """
        Client authentication can be set to use CHAP/CHAP_MUTUAL by supplying
        username, password, mutual_username, mutual_password

        e.g.
        auth username=<user> password=<pass> mutual_username=<m_user> mutual_password=<m_pass>

        username / mutual_username ... the username is 8-64 character string. Each character
                                       may either be an alphanumeric or use one of the following
                                       special characters .,:,-,@.
                                       Consider using the hosts 'shortname' or the initiators IQN
                                       value as the username

        password / mutual_password ... the password must be between 12-16 chars in length
                                       containing alphanumeric characters, plus the following
                                       special characters @,_,-,/

        WARNING1: Using unsupported special characters may result in truncation,
                  resulting in failed logins.

        WARNING2: If there are multiple clients, CHAP must be enabled for all
        clients or  disabled for all clients. gwcli does not support mixing CHAP
        clients with IQN ACL clients.

        """

        self.logger.debug("CMD: ../hosts/<client_iqn> auth *")

        if not username:
            self.logger.error("To set authentication, specify "
                              "username=<user> password=<password> "
                              "[mutual_username]=<user> [mutual_password]=<password> "
                              "or nochap")
            return

        self.set_auth(username, password, mutual_username, mutual_password)

    @staticmethod
    def get_srtd_names(lun_list):
        """
        sort the supplied list of luns (tuples - [('disk1',1),('disk2',0)])
        :return: list of LUN names, in lun_id sequence
        """

        srtd_luns = sorted(lun_list, key=lambda field: field[1])

        return [rbd_name for rbd_name, lun_id in srtd_luns]

    def ui_command_disk(self, action='add', disk=None, size=None):
        """
        Disks can be added or removed from the client one at a time using
        the 'disk' sub-command. Note that if the disk does not currently exist
        in the configuration, the cli will attempt to create it for you.

        e.g.
        disk add <pool_name/image_name> <size>
        disk remove <pool_name/image_name>

        Adding a disk will result in the disk occupying the client's next
        available lun id. Once allocated removing a LUN will not change the
        LUN id associations for the client.

        Note that if the client is a member of a host group, disk management
        *must* be performed at the group level. Attempting to add/remove disks
        at the client level will fail.

        """

        self.logger.debug("CMD: ../hosts/<client_iqn> disk action={}"
                          " disk={}".format(action,
                                            disk))

        valid_actions = ['add', 'remove']

        if not disk:
            self.logger.critical("You must supply a disk name to add/remove "
                                 "for this client")
            return

        if action not in valid_actions:
            self.logger.error("you can only add and remove disks - {} is "
                              "invalid ".format(action))
            return

        lun_list = [(lun.rbd_name, lun.lun_id) for lun in self.children]
        current_luns = Client.get_srtd_names(lun_list)

        if action == 'add':

            if disk not in current_luns:
                ui_root = self.get_ui_root()
                all_pools = ui_root.disks.children
                all_disks = []
                for current_pool in all_pools:
                    for current_disk in current_pool.children:
                        all_disks.append(current_disk)
                valid_disk_names = [defined_disk.image_id
                                    for defined_disk in all_disks]
            else:
                # disk provided is already mapped, so remind the user
                self.logger.error("Disk {} already mapped".format(disk))
                return
        else:
            valid_disk_names = current_luns

        if disk not in valid_disk_names:

            # if this is an add operation, we can create the disk on-the-fly
            # for the admin

            if action == 'add':
                ui_root = self.get_ui_root()
                ui_disks = ui_root.disks

                # a disk given here would be of the form pool.image
                try:
                    pool, image = disk.split('/')
                except ValueError:
                    self.logger.error("Invalid format. Use pool_name/disk_name")
                    return

                rc = ui_disks.create_disk(pool=pool, image=image, size=size)
                if rc == 0:
                    self.logger.debug("disk auto-define successful")
                else:
                    self.logger.error("disk auto-define failed({}), try "
                                      "using the /disks create "
                                      "command".format(rc))
                    return

            else:
                self.logger.error("disk '{}' is not mapped to this "
                                  "client ".format(disk))
                return

        mapped_disks = [mapped_disk.name
                        for mapped_disk in self.parent.parent.target_disks.children]
        if disk not in mapped_disks:
            rc = self.parent.parent.target_disks.add_disk(disk, None, None)
            if rc == 0:
                self.logger.debug("disk auto-map successful")
            else:
                self.logger.error("disk auto-map failed({}), try "
                                  "using the /iscsi-target/<iqn>/disks add "
                                  "command".format(rc))
                return

        # At this point we are either in add/remove mode, with a valid disk
        # to act upon
        self.logger.debug("Client '{}' update - {} disk "
                          "{}".format(self.client_iqn,
                                      action,
                                      disk))

        target_iqn = self.parent.parent.name

        api_vars = {"disk": disk}

        clientlun_api = ('{}://localhost:{}/api/'
                         'clientlun/{}/{}'.format(self.http_mode,
                                                  settings.config.api_port,
                                                  target_iqn,
                                                  self.client_iqn))

        api = APIRequest(clientlun_api, data=api_vars)
        if action == 'add':
            api.put()
        else:
            api.delete()

        if api.response.status_code == 200:

            self.logger.debug("disk mapping updated successfully")

            if action == 'add':

                # The addition of the lun will get a lun id assigned so
                # we need to query the api server to get the new configuration
                # to be able to set the local cli entry correctly
                get_api_vars = {"disk": disk}

                clientlun_api = clientlun_api.replace('/clientlun/',
                                                      '/_clientlun/')

                self.logger.debug("Querying API to get mapped LUN information")
                api = APIRequest(clientlun_api, data=get_api_vars)
                api.get()

                if api.response.status_code == 200:
                    try:
                        lun_dict = api.response.json()['message']
                    except Exception:
                        self.logger.error("Malformed REST API response")
                        return

                    # now update the UI
                    lun_id = lun_dict[disk]['lun_id']
                    self.add_lun(disk, lun_id)

                else:
                    self.logger.error("Query for disk '{}' meta data "
                                      "failed".format(disk))
                    return

            else:

                # this was a remove request, so simply delete the child
                # MappedLun object corresponding to this rbd name
                mlun = [lun for lun in self.children
                        if lun.rbd_name == disk][0]
                self.remove_lun(mlun)

            self.logger.debug("configuration update successful")
            self.logger.info('ok')

        else:
            # the request to add/remove the disk for the client failed
            self.logger.error("disk {} for '{}' against {} failed"
                              "\n{}".format(action,
                                            disk,
                                            self.client_iqn,
                                            response_message(api.response,
                                                             self.logger)))
            return

    def add_lun(self, disk, lun_id):

        MappedLun(self, disk, lun_id)

        # update the objects lun list (so ui info cmd picks
        # up the change
        self.luns[disk] = {'lun_id': lun_id}

        self.parent.update_lun_map('add',
                                   disk,
                                   self.client_iqn)

        active_maps = len(self.parent.lun_map[disk]) - 1
        if active_maps > 0:
            self.logger.warning("Warning: '{}' mapped to {} other "
                                "client(s)".format(disk,
                                                   active_maps))

    def remove_lun(self, lun):
        self.remove_child(lun)
        del self.luns[lun.rbd_name]
        self.parent.update_lun_map('remove',
                                   lun.rbd_name,
                                   self.client_iqn)

    @property
    def logged_in(self):
        target_iqn = self.parent.parent.name
        gateways = self.parent.parent.get_child('gateways')
        local_gw = this_host()
        is_local_target = len([child for child in gateways.children if child.name == local_gw]) > 0
        if is_local_target:
            client_info = GWClient.get_client_info(target_iqn, self.client_iqn)
            self.alias = client_info['alias']
            self.ip_address = ','.join(client_info['ip_address'])
            return client_info['state']
        else:
            self.alias = ''
            self.ip_address = ''
            return ''


class MappedLun(UINode):

    display_attributes = ["rbd_name", "owner", "size", "size_h", "lun_id"]

    def __init__(self, parent, name, lun_id):
        self.rbd_name = name
        UINode.__init__(self, 'lun {}'.format(lun_id), parent)

        # navigate back through the object model to pick up the disks
        ui_root = self.get_ui_root()
        disk_lookup = ui_root.disks.disk_lookup

        self.disk = disk_lookup[name]
        self.owner = self.disk.owner
        self.size = self.disk.size
        self.size_h = self.disk.size_h
        self.lun_id = lun_id

    def summary(self):
        self.owner = self.disk.owner
        self.size_h = self.disk.size_h
        return "{}({}), Owner: {}".format(self.rbd_name,
                                          self.size_h,
                                          self.owner), True



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json
import logging
import sys
import unittest

import mock

# We need to mock ceph libs python bindings because there's no
# updated package in pypy
sys.modules['rados'] = mock.Mock()
sys.modules['rbd'] = mock.Mock()

import ceph_iscsi_config.settings as settings  # noqa: E402
from ceph_iscsi_config.common import Config  # noqa: E402


class ChapTest(unittest.TestCase):

    def setUp(self):
        self.logger = logging.getLogger()
        settings.init()

    def test_upgrade_config(self):
        gateway_conf_initial = json.dumps(self.gateway_conf_initial)

        # First, the upgrade is executed on node1
        with mock.patch.object(Config, 'init_config', return_value=True), \
                mock.patch.object(Config, '_read_config_object',
                                  return_value=gateway_conf_initial), \
                mock.patch.object(Config, 'commit'), \
                mock.patch("socket.gethostname", return_value='node1'), \
                mock.patch("socket.getfqdn", return_value='node1.ceph.local'):
            config = Config(self.logger)

        # And then, the upgrade is executed on node2
        current_config = json.dumps(config.config)
        with mock.patch.object(Config, 'init_config', return_value=True), \
                mock.patch.object(Config, '_read_config_object',
                                  return_value=current_config), \
                mock.patch.object(Config, 'commit'), \
                mock.patch("socket.gethostname", return_value='node2'),\
                mock.patch("socket.getfqdn", return_value='node2.ceph.local'):
            config = Config(self.logger)

            self.maxDiff = None

            iqn = 'iqn.2003-01.com.redhat.iscsi-gw:iscsi-igw'
            self.assertGreater(config.config['targets'][iqn]['created'],
                               self.gateway_conf_latest['targets'][iqn]['created'])
            self.assertGreater(config.config['targets'][iqn]['updated'],
                               self.gateway_conf_latest['targets'][iqn]['updated'])
            config.config['targets'][iqn]['created'] = '2018/12/07 09:19:01'
            config.config['targets'][iqn]['updated'] = '2018/12/07 09:19:02'

            disk = 'rbd/disk_1'
            self.assertGreater(config.config['disks'][disk]['updated'],
                               self.gateway_conf_latest['disks'][disk]['updated'])
            config.config['disks'][disk]['updated'] = '2018/12/07 09:19:03'

            self.assertDictEqual(config.config, self.gateway_conf_latest)

    gateway_conf_initial = {
        "clients": {
            "iqn.1994-05.com.redhat:rh7-client": {
                "auth": {
                    "chap": "myiscsiusername/myiscsipassword"
                },
                "created": "2018/12/07 09:18:01",
                "group_name": "mygroup",
                "luns": {
                    "rbd.disk_1": {
                        "lun_id": 0
                    }
                },
                "updated": "2018/12/07 09:18:02"
            }
        },
        "controls": {
            "immediate_data": False,
            "nopin_response_timeout": 17
        },
        "created": "2018/12/07 09:18:03",
        "disks": {
            "rbd.disk_1": {
                "controls": {
                    "qfull_timeout": 18
                },
                "created": "2018/12/07 09:18:04",
                "image": "disk_1",
                "owner": "node1",
                "pool": "rbd",
                "pool_id": 7,
                "updated": "2018/12/07 09:18:05",
                "wwn": "4fc1071d-7e2f-4df0-95c8-925a617e2d62"
            }
        },
        "epoch": 19,
        "gateways": {
            "created": "2018/12/07 09:18:06",
            "ip_list": [
                "192.168.100.201",
                "192.168.100.202"
            ],
            "iqn": "iqn.2003-01.com.redhat.iscsi-gw:iscsi-igw",
            "node1": {
                "active_luns": 1,
                "created": "2018/12/07 09:18:07",
                "gateway_ip_list": [
                    "192.168.100.201",
                    "192.168.100.202"
                ],
                "inactive_portal_ips": [
                    "192.168.100.202"
                ],
                "iqn": "iqn.2003-01.com.redhat.iscsi-gw:iscsi-igw",
                "portal_ip_address": "192.168.100.201",
                "tpgs": 2,
                "updated": "2018/12/07 09:18:08"
            },
            "node2": {
                "active_luns": 0,
                "created": "2018/12/07 09:18:09",
                "gateway_ip_list": [
                    "192.168.100.201",
                    "192.168.100.202"
                ],
                "inactive_portal_ips": [
                    "192.168.100.201"
                ],
                "iqn": "iqn.2003-01.com.redhat.iscsi-gw:iscsi-igw",
                "portal_ip_address": "192.168.100.202",
                "tpgs": 2,
                "updated": "2018/12/07 09:18:10"
            }
        },
        "groups": {
            "mygroup": {
                "created": "2018/12/07 09:18:11",
                "disks": {
                    "rbd.disk_1": {
                        "lun_id": 0
                    }
                },
                "members": [
                    "iqn.1994-05.com.redhat:rh7-client"
                ],
                "updated": "2018/12/07 09:18:12"
            }
        },
        "updated": "2018/12/07 09:18:13",
        "version": 3
    }

    gateway_conf_latest = {
        "created": "2018/12/07 09:18:03",
        "disks": {
            "rbd/disk_1": {
                "controls": {
                    "qfull_timeout": 18
                },
                "created": "2018/12/07 09:18:04",
                "image": "disk_1",
                "owner": "node1.ceph.local",
                "backstore": "user:rbd",
                "backstore_object_name": "rbd.disk_1",
                "pool": "rbd",
                "pool_id": 7,
                "updated": "2018/12/07 09:19:03",
                "wwn": "4fc1071d-7e2f-4df0-95c8-925a617e2d62"
            }
        },
        "discovery_auth": {
            "username": "",
            "password": "",
            "password_encryption_enabled": False,
            "mutual_username": "",
            "mutual_password": "",
            "mutual_password_encryption_enabled": False
        },
        "epoch": 19,
        "gateways": {
            "node1.ceph.local": {
                "active_luns": 1,
                "created": "2018/12/07 09:18:07",
                "updated": "2018/12/07 09:18:08"
            },
            "node2.ceph.local": {
                "active_luns": 0,
                "created": "2018/12/07 09:18:09",
                "updated": "2018/12/07 09:18:10"
            }
        },
        "targets": {
            "iqn.2003-01.com.redhat.iscsi-gw:iscsi-igw": {
                "clients": {
                    "iqn.1994-05.com.redhat:rh7-client": {
                        "auth": {
                            "username": "myiscsiusername",
                            "password": "myiscsipassword",
                            "password_encryption_enabled": False,
                            "mutual_username": "",
                            "mutual_password": "",
                            "mutual_password_encryption_enabled": False
                        },
                        "group_name": "mygroup",
                        "luns": {
                            "rbd/disk_1": {
                                "lun_id": 0
                            }
                        }
                    }
                },
                "acl_enabled": True,
                "auth": {
                    "username": "",
                    "password": "",
                    "password_encryption_enabled": False,
                    "mutual_username": "",
                    "mutual_password": "",
                    "mutual_password_encryption_enabled": False
                },
                "controls": {
                    "immediate_data": False,
                    "nopin_response_timeout": 17
                },
                "created": "2018/12/07 09:19:01",
                "disks": {
                    "rbd/disk_1": {
                        "lun_id": 0
                    }
                },
                "groups": {
                    "mygroup": {
                        "disks": {
                            "rbd/disk_1": {
                                "lun_id": 0
                            }
                        },
                        "members": [
                            "iqn.1994-05.com.redhat:rh7-client"
                        ]
                    }
                },
                "ip_list": [
                    "192.168.100.201",
                    "192.168.100.202"
                ],
                "portals": {
                    "node1.ceph.local": {
                        "gateway_ip_list": [
                            "192.168.100.201",
                            "192.168.100.202"
                        ],
                        "inactive_portal_ips": [
                            "192.168.100.202"
                        ],
                        "portal_ip_addresses": ["192.168.100.201"],
                        "tpgs": 2
                    },
                    "node2.ceph.local": {
                        "gateway_ip_list": [
                            "192.168.100.201",
                            "192.168.100.202"
                        ],
                        "inactive_portal_ips": [
                            "192.168.100.201"
                        ],
                        "portal_ip_addresses": ["192.168.100.202"],
                        "tpgs": 2
                    }
                },
                "updated": "2018/12/07 09:19:02"
            }
        },
        "updated": "2018/12/07 09:18:13",
        "version": 11
    }



import requests
from requests import Response
import sys
import re
import os
import subprocess


from rtslib_fb.utils import normalize_wwn, RTSLibError

from ceph_iscsi_config.client import GWClient
import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import (resolve_ip_addresses, CephiSCSIError,
                                     this_host)

__author__ = 'Paul Cuzner'


class Colors(object):

    map = {'green': '\x1b[32;1m',
           'red': '\x1b[31;1m',
           'yellow': '\x1b[33;1m',
           'blue': '\x1b[34;1m'}


def readcontents(filename):
    with open(filename, 'r') as input_file:
        content = input_file.read().rstrip()
    return content


def get_config():
    """
    use the /config api to return the current gateway configuration
    :return: (dict) of the config object
    """

    http_mode = "https" if settings.config.api_secure else "http"
    api_rqst = "{}://localhost:{}/api/config".format(http_mode,
                                                     settings.config.api_port)
    api = APIRequest(api_rqst)
    api.get()

    if api.response.status_code == 200:
        try:
            return api.response.json()
        except Exception:
            pass

    return {}


def valid_gateway(target_iqn, gw_name, gw_ips, config):
    """
    validate the request for a new gateway
    :param gw_name: (str) host (shortname) of the gateway
    :param gw_ips: (str) ip addresses on the gw that will be used for iSCSI
    :param config: (dict) current config
    :return: (str) "ok" or error description
    """

    http_mode = 'https' if settings.config.api_secure else "http"

    # if the gateway request already exists in the config, computer says "no"
    target_config = config['targets'][target_iqn]
    if gw_name in target_config['portals']:
        return "Gateway name {} already defined".format(gw_name)

    for gw_ip in gw_ips:
        if gw_ip in target_config.get('ip_list', []):
            return "IP address already defined to the configuration"

    # validate the gateway name is resolvable
    if not resolve_ip_addresses(gw_name):
        return ("Gateway '{}' is not resolvable to an IP address".format(gw_name))

    # validate the ip_address is valid ip
    for gw_ip in gw_ips:
        if not resolve_ip_addresses(gw_ip):
            return ("IP address provided is not usable (name doesn't"
                    " resolve, or not a valid IPv4/IPv6 address)")

    # At this point the request seems reasonable, so lets check a bit deeper

    gw_api = '{}://{}:{}/api'.format(http_mode,
                                     gw_name,
                                     settings.config.api_port)

    # check the intended host actually has the requested IP available
    api = APIRequest(gw_api + '/sysinfo/ip_addresses')
    api.get()

    if api.response.status_code != 200:
        return ("ip_addresses query to {} failed - check "
                "rbd-target-api log. Is the API server "
                "running and in the right mode (http/https)?".format(gw_name))

    try:
        target_ips = api.response.json()['data']
    except Exception:
        return "Malformed REST API response"

    for gw_ip in gw_ips:
        if gw_ip not in target_ips:
            return ("IP address of {} is not available on {}. Valid "
                    "IPs are :{}".format(gw_ip,
                                         gw_name,
                                         ','.join(target_ips)))

    # check that config file on the new gateway matches the local machine
    api = APIRequest(gw_api + '/sysinfo/checkconf')
    api.get()
    if api.response.status_code != 200:
        return ("checkconf API call to {} failed with "
                "code".format(gw_name,
                              api.response.status_code))

    # compare the hash of the new gateways conf file with the local one
    local_hash = settings.config.hash()
    try:
        remote_hash = str(api.response.json()['data'])
    except Exception:
        remote_hash = None

    if local_hash != remote_hash:
        return ("/etc/ceph/iscsi-gateway.cfg on {} does "
                "not match the local version. Correct and "
                "retry request".format(gw_name))

    # Check for package version dependencies
    api = APIRequest(gw_api + '/sysinfo/checkversions')
    api.get()
    if api.response.status_code != 200:
        try:
            errors = api.response.json()['data']
        except Exception:
            return "Malformed REST API response"

        return ("{} failed package validation checks - "
                "{}".format(gw_name,
                            ','.join(errors)))

    # At this point the gateway seems valid
    return "ok"


def get_remote_gateways(config, logger, local_gw_required=True):
    """
    Return the list of remote gws.
    :param: config: Config object with gws setup.
    :param: logger: Logger object
    :param: local_gw_required: Check if local_gw is defined within gateways configuration
    :return: A list of gw names, or CephiSCSIError if not run on a gw in the
             config
    """

    local_gw = this_host()
    logger.debug("this host is {}".format(local_gw))
    gateways = [key for key in config
                if isinstance(config[key], dict)]
    logger.debug("all gateways - {}".format(gateways))
    if local_gw_required and local_gw not in gateways:
        raise CephiSCSIError("{} cannot be used to perform this operation "
                             "because it is not defined within the gateways "
                             "configuration".format(local_gw))
    if local_gw in gateways:
        gateways.remove(local_gw)
    logger.debug("remote gateways: {}".format(gateways))
    return gateways


def valid_credentials(username, password, mutual_username, mutual_password):
    """
    Returns `None` if credentials are acceptable, otherwise return an error message

    username / mutual_username is 8-64 chars long containing any alphanumeric in
    [0-9a-zA-Z] and '.' ':' '@' '_' '-'

    password / mutual_password is 12-16 chars long containing any alphanumeric in
    [0-9a-zA-Z] and '@' '-' '_' '/'
    """

    usr_regex = re.compile(r"^[\w\\.\:\@\_\-]{8,64}$")
    pw_regex = re.compile(r"^[\w\@\-\_\/]{12,16}$")

    if username and not password:
        return 'Password is required'

    if not username and (password or mutual_username):
        return 'Username is required'

    if mutual_username and not mutual_password:
        return 'Mutual password is required'

    if not mutual_username and mutual_password:
        return 'Mutual username is required'

    if username and len(username) < 8:
        return 'Minimum length of username is 8 characters'

    if username and len(username) > 64:
        return 'Maximum length of username is 64 characters'

    if username and not usr_regex.search(username):
        return 'Invalid username'

    if mutual_username and len(mutual_username) < 8:
        return 'Minimum length of mutual username is 8 characters'

    if mutual_username and len(mutual_username) > 64:
        return 'Maximum length of mutual username is 64 characters'

    if mutual_username and not usr_regex.search(mutual_username):
        return 'Invalid mutual username'

    if password and len(password) < 12:
        return 'Minimum length of password is 12 characters'

    if password and len(password) > 16:
        return 'Maximum length of password is 16 characters'

    if password and not pw_regex.search(password):
        return 'Invalid password'

    if mutual_password and len(mutual_password) < 12:
        return 'Minimum length of mutual password is 12 characters'

    if mutual_password and len(mutual_password) > 16:
        return 'Maximum length of mutual password is 16 characters'

    if mutual_password and not pw_regex.search(mutual_password):
        return 'Invalid mutual password'

    return None


def valid_client(**kwargs):
    """
    validate a client create or update request, based on mode.
    :param kwargs: 'mode' is the key field used to determine process flow
    :return: 'ok' or an error description (str)
    """

    valid_modes = ['create', 'delete', 'auth', 'disk']
    parms_passed = set(kwargs.keys())

    if 'mode' in kwargs:
        if kwargs['mode'] not in valid_modes:
            return ("Invalid client validation mode request - "
                    "asked for {}, available {}".format(kwargs['mode'],
                                                        valid_modes))
    else:
        return "Invalid call to valid_client - mode is needed"

    # at this point we have a mode to work with

    mode = kwargs['mode']
    client_iqn = kwargs['client_iqn']
    target_iqn = kwargs['target_iqn']
    config = get_config()
    if not config:
        return "Unable to query the local API for the current config"
    target_config = config['targets'][target_iqn]

    if mode == 'create':
        # iqn must be valid
        try:
            normalize_wwn(['iqn'], client_iqn)
        except RTSLibError:
            return ("Invalid IQN name for iSCSI")

        # iqn must not already exist
        if client_iqn in target_config['clients']:
            return ("A client with the name '{}' is "
                    "already defined".format(client_iqn))

        # Mixing TPG/target auth with ACL is not supported
        target_username = target_config['auth']['username']
        target_password = target_config['auth']['password']
        target_auth_enabled = (target_username and target_password)
        if target_auth_enabled:
            return "Cannot create client because target CHAP authentication is enabled"

        # Creates can only be done with a minimum number of gw's in place
        num_gws = len([gw_name for gw_name in config['gateways']
                       if isinstance(config['gateways'][gw_name], dict)])
        if num_gws < settings.config.minimum_gateways:
            return ("Clients can not be defined until a HA configuration "
                    "has been defined "
                    "(>{} gateways)".format(settings.config.minimum_gateways))

        # at this point pre-req's look good
        return 'ok'

    elif mode == 'delete':

        # client must exist in the configuration
        if client_iqn not in target_config['clients']:
            return ("{} is not defined yet - nothing to "
                    "delete".format(client_iqn))

        this_client = target_config['clients'].get(client_iqn)
        if this_client.get('group_name', None):
            return ("Unable to delete '{}' - it belongs to "
                    "group {}".format(client_iqn,
                                      this_client.get('group_name')))

        # client to delete must not be logged in - we're just checking locally,
        # since *all* nodes are set up the same, and a client login request
        # would normally login to each gateway
        client_info = GWClient.get_client_info(target_iqn, client_iqn)
        if client_info['state'] == 'LOGGED_IN':
            return ("Client '{}' is logged in to {}- unable to delete until"
                    " it's logged out".format(client_iqn, target_iqn))

        # at this point, the client looks ok for a DELETE operation
        return 'ok'

    elif mode == 'auth':
        # client iqn must exist
        if client_iqn not in target_config['clients']:
            return ("Client '{}' does not exist".format(client_iqn))

        username = kwargs['username']
        password = kwargs['password']
        mutual_username = kwargs['mutual_username']
        mutual_password = kwargs['mutual_password']

        error_msg = valid_credentials(username, password, mutual_username, mutual_password)
        if error_msg:
            return error_msg

        return 'ok'

    elif mode == 'disk':

        this_client = target_config['clients'].get(client_iqn)
        if this_client.get('group_name', None):
            return ("Unable to manage disks for '{}' - it belongs to "
                    "group {}".format(client_iqn,
                                      this_client.get('group_name')))

        if 'image_list' not in parms_passed:
            return ("Disk changes require 'image_list' to be set, containing"
                    " a comma separated str of rbd images (pool/image)")

        rqst_disks = set(kwargs['image_list'].split(','))
        mapped_disks = set(target_config['clients'][client_iqn]['luns'].keys())
        current_disks = set(config['disks'].keys())

        if len(rqst_disks) > len(mapped_disks):
            # this is an add operation

            # ensure the image list is 'complete' not just a single disk
            if not mapped_disks.issubset(rqst_disks):
                return ("Invalid image list - it must contain existing "
                        "disks AND any additions")

            # ensure new disk(s) exist - must yield a result since rqst>mapped
            new_disks = rqst_disks.difference(mapped_disks)
            if not new_disks.issubset(current_disks):
                # disks provided are not currently defined
                return ("Invalid image list - it defines new disks that do "
                        "not current exist")

            return 'ok'

        else:

            # this is a disk removal operation
            if kwargs['image_list']:
                if not rqst_disks.issubset(mapped_disks):
                    return ("Invalid image list ({})".format(rqst_disks))

            return 'ok'

    return 'Unknown error in valid_client function'


def valid_snapshot_name(name):
    regex = re.compile("^[^/@]+$")
    if not regex.search(name):
        return False
    return True


def refresh_control_values(control_values, controls, def_settings):
    for key, setting in def_settings.items():
        val = controls.get(setting.name)
        if val is not None:
            # config values may be normalized or raw
            val = setting.to_str(setting.normalize(val))

        def_val = setting.to_str(getattr(settings.config, key))

        if val is None or val == def_val:
            control_values[setting.name] = def_val
        else:
            control_values[setting.name] = "{} (override)".format(val)


class GatewayError(Exception):
    pass


class GatewayAPIError(GatewayError):
    pass


class GatewayLIOError(GatewayError):
    pass


class APIRequest(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        # Establish defaults for the API connection
        if 'auth' not in self.kwargs:
            self.kwargs['auth'] = (settings.config.api_user,
                                   settings.config.api_password)
        if 'verify' not in self.kwargs:
            self.kwargs['verify'] = settings.config.api_ssl_verify

        self.http_methods = ['get', 'put', 'delete']
        self.data = None

    def _get_response(self):
        return self.data

    def __getattr__(self, name):
        if name in self.http_methods:
            request_method = getattr(requests, name)
            try:
                self.data = request_method(*self.args, **self.kwargs)
            except requests.ConnectionError:
                msg = ("Unable to connect to api endpoint @ "
                       "{}".format(self.args[0]))
                self.data = Response()
                self.data.status_code = 500
                self.data._content = '{{"message": "{}" }}'.format(msg)
                return self._get_response
            except Exception:
                raise GatewayAPIError("Unknown error connecting to "
                                      "{}".format(self.args[0]))
            else:
                # since the attribute is a callable, we must return with
                # a callable
                return self._get_response
        raise AttributeError()

    response = property(_get_response,
                        doc="get http response output")


def progress_message(text, color='green'):

    sys.stdout.write("{}{}{}\r".format(Colors.map[color],
                                       text,
                                       '\x1b[0m'))
    sys.stdout.flush()


def console_message(text, color='green'):

    color_needed = getattr(settings.config, 'interactive', True)

    if color_needed:
        print("{}{}{}".format(Colors.map[color],
                              text,
                              '\x1b[0m'))
    else:
        print(text)


def cmd_exists(command):
    return any(
        os.access(os.path.join(path, command), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )


def os_cmd(command):
    """
    Issue a command to the OS and return the output. NB. check_output default
    is shell=False
    :param command: (str) OS command
    :return: (str) command response (lines terminated with \n)
    """
    cmd_list = command.split(' ')
    if cmd_exists(cmd_list[0]):
        cmd_output = subprocess.check_output(cmd_list,
                                             stderr=subprocess.STDOUT).rstrip()
        return cmd_output
    else:
        return ''


def response_message(response, logger=None):
    """
    Attempts to retrieve the "message" value from a JSON-encoded response
    message. If the JSON fails to parse, the response will be returned
    as-is.
    :param response: (requests.Response) response
    :param logger: optional logger
    :return: (str) response message
    """
    try:
        return response.json()['message']
    except Exception:
        if logger:
            logger.debug("Failed API request: {} {}\n{}".format(response.request.method,
                                                                response.request.url,
                                                                response.text))
        return "{} {}".format(response.status_code, response.reason)



# import ceph_iscsi_config.settings as settings
import json

from ceph_iscsi_config.common import Config
from ceph_iscsi_config.client import GWClient
from ceph_iscsi_config.utils import ListComparison


class Group(object):

    def __init__(self, logger, target_iqn, group_name, members=[], disks=[]):

        """
        Manage a host group definition. The input for the group object is the
        desired state of the group where the logic enforced produces an
        idempotent group definition across API/CLI and more importantly Ansible

        :param logger: (logging object) used for centralised logging
        :param target_iqn: (str) target iqn
        :param group_name: (str) group name
        :param members: (list) iscsi IQN's of the clients
        :param disks: (list) disk names of the format pool/image
        """

        self.logger = logger

        self.error = False
        self.error_msg = ''
        self.num_changes = 0

        self.config = Config(logger)
        if self.config.error:
            self.error = self.config.error
            self.error_msg = self.config.error_msg
            return

        self.target_iqn = target_iqn
        self.group_name = group_name
        self.group_members = members
        self.disks = disks

        target_config = self.config.config['targets'][self.target_iqn]
        if group_name in target_config['groups']:
            self.new_group = False
        else:
            self.new_group = True

        self.logger.debug("Group : name={}".format(self.group_name))
        self.logger.debug("Group : members={}".format(self.group_members))
        self.logger.debug("Group : disks={}".format(self.disks))

    def __str__(self):
        return ("Group: {}\n- Members: {}\n- "
                "Disks: {}".format(self.group_name,
                                   self.group_members,
                                   self.disks))

    def _set_error(self, error_msg):
        self.error = True
        self.error_msg = error_msg
        self.logger.debug("Error: {}".format(self.error_msg))

    def _valid_client(self, action, client_iqn):
        """
        validate the addition of a specific client
        :param action: (str) add or remove request
        :param client_iqn: (str) iqn of the client to add tot he group
        :return: (bool) true/false whether the client should be accepted
        """

        target_config = self.config.config['targets'][self.target_iqn]

        self.logger.debug("checking '{}'".format(client_iqn))

        # to validate the request, pass through a 'negative' filter
        if action == 'add':
            client = target_config['clients'].get(client_iqn, {})
            if not client:
                self._set_error("client '{}' doesn't exist".format(client_iqn))
                return False
            elif client.get('luns'):
                self._set_error("Client '{}' already has luns. "
                                "Only clients without prior lun maps "
                                "can be added to a group".format(client_iqn))
                return False
            elif client.get('group_name'):
                self._set_error("Client already assigned to {} - a client "
                                "can only belong to one host "
                                "group".format(client.get('group_name')))
                return False
        else:
            # client_iqn must exist in the group
            if client_iqn not in target_config['groups'][self.group_name].get('members'):
                self._set_error("client '{}' is not a member of "
                                "{}".format(client_iqn,
                                            self.group_name))
                return False

        # to reach here the request is considered valid
        self.logger.debug("'{}' client '{}' for group '{}'"
                          " is valid".format(action,
                                             client_iqn,
                                             self.group_name))
        return True

    def _valid_disk(self, action, disk):

        self.logger.debug("checking disk '{}'".format(disk))
        target_config = self.config.config['targets'][self.target_iqn]
        if action == 'add':

            if disk not in target_config['disks']:
                self._set_error("disk '{}' doesn't exist".format(disk))
                return False
        else:
            if disk not in target_config['groups'][self.group_name]['disks']:
                self._set_error("disk '{}' is not in the group".format(disk))
                return False

        return True

    def _next_lun(self, preferred_lun_id):
        """
        Look at the disk list for the group and return the 1st available free
        LUN id used for adding disks to the group
        :return: (int) lun Id
        """

        lun_range = list(range(0, 256, 1))  # 0->255
        lun_range.remove(preferred_lun_id)
        lun_range.insert(0, preferred_lun_id)

        target_config = self.config.config['targets'][self.target_iqn]
        group = target_config['groups'][self.group_name]
        group_disks = group.get('disks')
        for d in group_disks:
            lun_range.remove(group_disks[d].get('lun_id'))

        return lun_range[0]

    def apply(self):
        """
        setup/manage the group definition
        :return: NULL
        """
        group_seed = {
            "members": [],
            "disks": {}
        }

        target_config = self.config.config['targets'][self.target_iqn]

        if self.new_group:

            # New Group definition, so seed it
            self.logger.debug("Processing request for new group "
                              "'{}'".format(self.group_name))
            if len(set(self.group_members)) != len(self.group_members):
                self._set_error("Member must contain unique clients - no "
                                "duplication")
                return

            self.logger.debug("New group definition required")

            # new_group = True
            target_config['groups'][self.group_name] = group_seed

        # Now the group definition is at least seeded, so let's look at the
        # member and disk information passed

        this_group = target_config['groups'][self.group_name]

        members = ListComparison(this_group.get('members'),
                                 self.group_members)
        disks = ListComparison(list(this_group.get('disks').keys()),
                               self.disks)

        if set(self.disks) != set(this_group.get('disks')) or \
                set(self.group_members) != set(this_group.get('members')):
            group_changed = True
        else:
            group_changed = False

        if group_changed or self.new_group:

            if self.valid_request(members, disks):
                self.update_metadata(members, disks)
            else:
                self._set_error("Group request failed validation")
                return

        else:
            # no changes required
            self.logger.info("Current group definition matches request")

        self.enforce_policy()

    def valid_request(self, members, disks):

        self.logger.info("Validating client membership")
        for mbr in members.added:
            if not self._valid_client('add', mbr):
                self.logger.error("'{}' failed checks".format(mbr))
                return False
        for mbr in members.removed:
            if not self._valid_client('remove', mbr):
                self.logger.error("'{}' failed checks".format(mbr))
                return False

        self.logger.debug("Client membership checks passed")
        self.logger.debug("clients to add : {}".format(members.added))
        self.logger.debug("clients to remove : {}".format(members.removed))

        # client membership is valid, check disks
        self.logger.info("Validating disk membership")
        for disk_name in disks.added:
            if not self._valid_disk('add', disk_name):
                self.logger.error("'{}' failed checks".format(disk_name))
                return False
        for disk_name in disks.removed:
            if not self._valid_disk('remove', disk_name):
                self.logger.error("'{}' failed checks".format(disk_name))
                return False

        self.logger.info("Disk membership checks passed")
        self.logger.debug("disks to add : {}".format(disks.added))
        self.logger.debug("disks to remove : {}".format(disks.removed))

        return True

    def update_metadata(self, members, disks):

        target_config = self.config.config['targets'][self.target_iqn]
        this_group = target_config['groups'].get(self.group_name, {})
        group_disks = this_group.get('disks', {})
        if disks.added:
            # update the groups disk list
            for disk in disks.added:
                lun_seq = self._next_lun(target_config['disks'][disk]['lun_id'])
                group_disks[disk] = {"lun_id": lun_seq}
                self.logger.debug("- adding '{}' to group '{}' @ "
                                  "lun id {}".format(disk,
                                                     self.group_name,
                                                     lun_seq))

        if disks.removed:
            # remove disk from the group definition
            for disk in disks.removed:
                del group_disks[disk]
                self.logger.debug("- removed '{}' from group "
                                  "{}".format(disk,
                                              self.group_name))

        if disks.added or disks.removed:
            # update each clients meta data
            self.logger.debug("updating clients LUN masking with "
                              "{}".format(json.dumps(group_disks)))

            for client_iqn in self.group_members:
                self.update_disk_md(client_iqn, group_disks)

        # handle client membership
        if members.changed:
            for client_iqn in members.added:
                self.add_client(client_iqn)
                self.update_disk_md(client_iqn, group_disks)
            for client_iqn in members.removed:
                self.remove_client(client_iqn)

        this_group['members'] = self.group_members
        this_group['disks'] = group_disks

        self.logger.debug("Group '{}' updated to "
                          "{}".format(self.group_name,
                                      json.dumps(this_group)))
        target_config['groups'][self.group_name] = this_group
        self.config.update_item('targets', self.target_iqn, target_config)
        self.config.commit()

    def enforce_policy(self):

        target_config = self.config.config['targets'][self.target_iqn]
        this_group = target_config['groups'][self.group_name]
        group_disks = this_group.get('disks')
        host_group = this_group.get('members')

        image_list = sorted(group_disks.items(),
                            key=lambda v: v[1]['lun_id'])

        for client_iqn in host_group:
            self.update_client(client_iqn, image_list)
            if self.error:
                # Applying the policy failed, so report and abort
                self.logger.error("Unable to apply policy to {} "
                                  ": {}".format(client_iqn,
                                                self.error_msg))
                return

    def add_client(self, client_iqn):
        target_config = self.config.config['targets'][self.target_iqn]
        client_metadata = target_config['clients'][client_iqn]
        client_metadata['group_name'] = self.group_name
        self.config.update_item('targets', self.target_iqn, target_config)
        self.logger.info("Added {} to group {}".format(client_iqn,
                                                       self.group_name))

    def update_disk_md(self, client_iqn, group_disks):
        target_config = self.config.config['targets'][self.target_iqn]
        md = target_config['clients'].get(client_iqn)
        md['luns'] = group_disks
        self.config.update_item('targets', self.target_iqn, target_config)
        self.logger.info("updated {} disk map to "
                         "{}".format(client_iqn,
                                     json.dumps(group_disks)))

    def update_client(self, client_iqn, image_list):

        client = GWClient(self.logger, client_iqn, image_list, '', '', '', '', self.target_iqn)
        client.manage('reconfigure')

        # grab the client's metadata from the config (needed by setup_luns)
        target_config = self.config.config['targets'][self.target_iqn]
        client.metadata = target_config['clients'][client_iqn]
        client.setup_luns(self.config.config['disks'])

        if client.error:
            self._set_error(client.error_msg)

    def remove_client(self, client_iqn):
        target_config = self.config.config['targets'][self.target_iqn]
        client_md = target_config["clients"][client_iqn]

        # remove the group_name setting from the client
        client_md['group_name'] = ''
        self.config.update_item('targets', self.target_iqn, target_config)
        self.logger.info("Removed {} from group {}".format(client_iqn,
                                                           self.group_name))

    def purge(self):

        # act on the group name
        # get the members from the current definition
        target_config = self.config.config['targets'][self.target_iqn]
        groups = target_config['groups']
        if self.group_name in groups:
            for mbr in groups[self.group_name]["members"]:
                self.remove_client(mbr)

            # issue a del_item to the config object for this group_name
            groups.pop(self.group_name)
            self.config.update_item('targets', self.target_iqn, target_config)
            self.config.commit()
            self.logger.info("Group {} removed".format(self.group_name))
        else:

            self._set_error("Group name requested does not exist")
            return



import subprocess
import netifaces

from rtslib_fb.utils import RTSLibError
from rtslib_fb.fabric import ISCSIFabricModule
from rtslib_fb.target import Target

import ceph_iscsi_config.settings as settings

from ceph_iscsi_config.target import GWTarget
from ceph_iscsi_config.lun import LUN
from ceph_iscsi_config.client import GWClient
from ceph_iscsi_config.lio import LIO
from ceph_iscsi_config.utils import this_host, CephiSCSIError

__author__ = 'pcuzner@redhat.com'


class CephiSCSIGateway(object):

    def __init__(self, logger, config, name=None):
        self.logger = logger
        self.config = config
        if name:
            self.hostname = name
        else:
            self.hostname = this_host()

    def ceph_rm_blacklist(self, blacklisted_ip):
        """
        Issue a ceph osd blacklist rm command for a given IP on this host
        :param blacklisted_ip: IP address (str - dotted quad)
        :return: boolean for success of the rm operation
        """

        self.logger.info("Removing blacklisted entry for this host : "
                         "{}".format(blacklisted_ip))

        conf = settings.config
        result = subprocess.check_output("ceph -n {client_name} --conf {cephconf} "
                                         "osd blacklist rm {blacklisted_ip}".
                                         format(blacklisted_ip=blacklisted_ip,
                                                client_name=conf.cluster_client_name,
                                                cephconf=conf.cephconf),
                                         stderr=subprocess.STDOUT, shell=True)
        if ("un-blacklisting" in result) or ("isn't blacklisted" in result):
            self.logger.info("Successfully removed blacklist entry")
            return True
        else:
            self.logger.critical("blacklist removal failed. Run"
                                 " 'ceph -n {client_name} --conf {cephconf} "
                                 "osd blacklist rm {blacklisted_ip}'".
                                 format(blacklisted_ip=blacklisted_ip,
                                        client_name=conf.cluster_client_name,
                                        cephconf=conf.cephconf))
            return False

    def osd_blacklist_cleanup(self):
        """
        Process the osd's to see if there are any blacklist entries for this
        node
        :return: True, blacklist entries removed OK, False - problems removing
        a blacklist
        """

        self.logger.info("Processing osd blacklist entries for this node")

        cleanup_state = True
        conf = settings.config

        try:

            # NB. Need to use the stderr override to catch the output from
            # the command
            blacklist = subprocess.check_output("ceph -n {client_name} --conf {cephconf} "
                                                "osd blacklist ls"
                                                .format(client_name=conf.cluster_client_name,
                                                        cephconf=conf.cephconf),
                                                shell=True,
                                                stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            self.logger.critical("Failed to run 'ceph -n {client_name} --conf {cephconf} "
                                 "osd blacklist ls'. Please resolve manually..."
                                 .format(client_name=conf.cluster_client_name,
                                         cephconf=conf.cephconf))
            cleanup_state = False
        else:

            blacklist_output = blacklist.decode('utf-8').split('\n')[:-1]
            if len(blacklist_output) > 1:

                # We have entries to look for, so first build a list of ipv4
                # addresses on this node
                ipv4_list = []
                for iface in netifaces.interfaces():
                    dev_info = netifaces.ifaddresses(iface).get(netifaces.AF_INET, [])
                    ipv4_list += [dev['addr'] for dev in dev_info]

                # process the entries (first entry just says "Listed X entries,
                # last entry is just null)
                for blacklist_entry in blacklist_output[1:]:

                    # valid entries to process look like -
                    # 192.168.122.101:0/3258528596 2016-09-28 18:23:15.307227
                    blacklisted_ip = blacklist_entry.split(':')[0]
                    # Look for this hosts ipv4 address in the blacklist

                    if blacklisted_ip in ipv4_list:
                        # pass in the ip:port/nonce
                        rm_ok = self.ceph_rm_blacklist(blacklist_entry.split(' ')[0])
                        if not rm_ok:
                            cleanup_state = False
                            break
            else:
                self.logger.info("No OSD blacklist entries found")

        return cleanup_state

    def get_tpgs(self, target_iqn):
        """
        determine the number of tpgs in the current target
        :return: count of the defined tpgs
        """

        try:
            target = Target(ISCSIFabricModule(), target_iqn, "lookup")

            return len([tpg.tag for tpg in target.tpgs])
        except RTSLibError:
            return 0

    def portals_active(self, target_iqn):
        """
        use the get_tpgs function to determine whether there are tpg's defined
        :return: (bool) indicating whether there are tpgs defined
        """
        return self.get_tpgs(target_iqn) > 0

    def redefine_target(self, target_iqn):
        self.delete_target(target_iqn)

        target_config = self.config.config['targets'][target_iqn]
        self.define_target(target_iqn, target_config['ip_list'])

    def define_target(self, target_iqn, gw_ip_list, target_only=False):
        """
        define the iSCSI target and tpgs
        :param target_iqn: (str) target iqn
        :param gw_ip_list: (list) gateway ip list
        :param target_only: (bool) if True only setup target
        :return: (object) GWTarget object
        """

        # GWTarget Definition : Handle the creation of the Target/TPG(s) and
        # Portals. Although we create the tpgs, we flick the enable_portal flag
        # off so the enabled tpg will not have an outside IP address. This
        # prevents clients from logging in too early, failing and giving up
        # because the nodeACL hasn't been defined yet (yes Windows I'm looking
        # at you!)

        # first check if there are tpgs already in LIO (True) - this would
        # indicate a restart or reload call has been made. If the tpg count is
        # 0, this is a boot time request
        self.logger.info("Setting up {}".format(target_iqn))

        target = GWTarget(self.logger, target_iqn, gw_ip_list,
                          enable_portal=self.portals_active(target_iqn))
        if target.error:
            raise CephiSCSIError("Error initializing iSCSI target: "
                                 "{}".format(target.error_msg))

        target.manage('target')
        if target.error:
            raise CephiSCSIError("Error creating the iSCSI target (target, "
                                 "TPGs, Portals): {}".format(target.error_msg))

        if not target_only:
            self.logger.info("Processing LUN configuration")
            try:
                LUN.define_luns(self.logger, self.config, target)
            except CephiSCSIError as err:
                self.logger.error("{} - Could not define LUNs: "
                                  "{}".format(target.iqn, err))
                raise

            self.logger.info("{} - Processing client configuration".
                             format(target.iqn))
            try:
                GWClient.define_clients(self.logger, self.config, target.iqn)
            except CephiSCSIError as err:
                self.logger.error("Could not define clients: {}".format(err))
                raise

        if not target.enable_portal:
            # The tpgs, luns and clients are all defined, but the active tpg
            # doesn't have an IP bound to it yet (due to the
            # enable_portals=False setting above)
            self.logger.info("{} - Adding the IP to the enabled tpg, "
                             "allowing iSCSI logins".format(target.iqn))
            target.enable_active_tpg(self.config)
            if target.error:
                raise CephiSCSIError("{} - Error enabling the IP with the "
                                     "active TPG: {}".format(target.iqn,
                                                             target.error_msg))
        return target

    def define_targets(self):
        """
        define the list of iSCSI targets and tpgs
        :return: (list) GWTarget objects
        """
        targets = []
        for iqn, target in self.config.config['targets'].items():
            if self.hostname in target['portals']:
                target = self.define_target(iqn, target.get('ip_list', {}))
                targets.append(target)
        return targets

    def define(self):
        """
        procesing logic that orchestrates the creation of the iSCSI gateway
        to LIO.
        """

        self.logger.info("Reading the configuration object to update local LIO "
                         "configuration")

        # first check to see if we have any entries to handle - if not, there is
        # no work to do..
        if "targets" not in self.config.config:
            self.logger.info("Configuration is empty - nothing to define to LIO")
            return

        if self.hostname not in self.config.config['gateways']:
            self.logger.info("Configuration does not have an entry for this host({}) - "
                             "nothing to define to LIO".format(self.hostname))
            return

        # at this point we have a gateway entry that applies to the running host

        self.logger.info("Processing Gateway configuration")
        self.define_targets()
        self.logger.info("Ceph iSCSI Gateway configuration load complete")

    def delete_target(self, target_iqn):

        target = GWTarget(self.logger, target_iqn, {})
        if target.error:
            raise CephiSCSIError("Could not initialize target: {}".
                                 format(target.error_msg))

        target.load_config()
        if target.error:
            self.logger.debug("Could not find target {}: {}".
                              format(target_iqn, target.error_msg))
            # Target might not be setup on this node. Ignore.
            return

        try:
            target.delete(self.config)
        except RTSLibError as err:
            err_msg = "Could not remove target {}: {}".format(target_iqn, err)
            raise CephiSCSIError(err_msg)

    def delete_targets(self):

        err_msg = None

        if self.hostname not in self.config.config['gateways']:
            return

        # Clear the current config, based on the config objects settings.
        # This will fail incoming IO, but wait on outstanding IO to
        # complete normally. We rely on the initiator multipath layer
        # to handle retries like a normal path failure.
        self.logger.info("Removing iSCSI target from LIO")

        for target_iqn, target_config in self.config.config['targets'].items():
            try:
                self.delete_target(target_iqn)
            except CephiSCSIError as err:
                if err_msg is None:
                    err_msg = err
                continue

        if err_msg:
            raise CephiSCSIError(err_msg)

    def delete(self):
        """
        Clear the LIO configuration of the settings defined by the config object
        We could simply call the clear_existing method of rtsroot - but if the
        admin has defined additional non ceph iscsi exports they'd loose
        everything

        :return: (int) 0 = LIO configuration removed/not-required
                       4 = LUN removal problem encountered
                       8 = Gateway (target/tpgs) removal failed
        """

        self.logger.debug("delete received, refreshing local state")
        self.config.refresh()
        if self.config.error:
            self.logger.critical("Problems accessing config object"
                                 " - {}".format(self.config.error_msg))
            return 8

        if "gateways" in self.config.config:
            if self.hostname not in self.config.config["gateways"]:
                self.logger.info("No gateway configuration to remove on this "
                                 "host ({})".format(self.hostname))
                return 0
        else:
            self.logger.info("Configuration object does not hold any gateway "
                             "metadata - nothing to do")
            return 0

        ret = 0

        try:
            self.delete_targets()
        except CephiSCSIError:
            ret = 8

        # unload disks not yet added to targets
        lio = LIO()
        lio.drop_lun_maps(self.config, False)
        if lio.error:
            self.logger.error("failed to remove LUN objects")
            if ret != 0:
                ret = 4

        if ret == 0:
            self.logger.info("Active Ceph iSCSI gateway configuration removed")
        return ret

    def remove_from_config(self, target_iqn):
        has_changed = False

        target_config = self.config.config["targets"].get(target_iqn)
        if target_config:
            local_gw = target_config['portals'].get(self.hostname)
            if local_gw:
                local_gw_ips = local_gw['portal_ip_addresses']

                target_config['portals'].pop(self.hostname)

                ip_list = target_config['ip_list']
                for local_gw_ip in local_gw_ips:
                    ip_list.remove(local_gw_ip)

                for _, remote_gw_config in target_config['portals'].items():
                    for local_gw_ip in local_gw_ips:
                        remote_gw_config["gateway_ip_list"].remove(local_gw_ip)
                        remote_gw_config["inactive_portal_ips"].remove(local_gw_ip)
                    tpg_count = remote_gw_config["tpgs"]
                    remote_gw_config["tpgs"] = tpg_count - 1

                if not target_config['portals']:
                    # Last gw for the target so delete everything that lives
                    # under the tpg in LIO since we can't create it
                    target_config['disks'] = {}
                    target_config['clients'] = {}
                    target_config['controls'] = {}
                    target_config['groups'] = {}

                has_changed = True
                self.config.update_item('targets', target_iqn, target_config)

        remove_gateway = True
        for _, target in self.config.config["targets"].items():
            if self.hostname in target['portals']:
                remove_gateway = False
                break

        if remove_gateway:
            # gateway is no longer used, so delete it
            has_changed = True
            self.config.del_item('gateways', self.hostname)

        LUN.reassign_owners(self.logger, self.config)

        if has_changed:
            self.config.commit("retain")
            if self.config.error:
                raise CephiSCSIError(self.config.error_msg)



from configshell_fb import ConfigNode
from gwcli.utils import console_message
import logging

__author__ = 'Paul Cuzner'


class UICommon(ConfigNode):

    def __init__(self, name, parent=None, shell=None):
        ConfigNode.__init__(self, name, parent, shell)
        self.logger = logging.getLogger('gwcli')

    def ui_command_goto(self, shortcut='/'):
        '''
        cd to the bookmark at shortcut.

        See 'help bookmarks' for more info on bookmarks.
        '''
        if shortcut in self.shell.prefs['bookmarks']:
            return self.ui_command_cd(self.shell.prefs['bookmarks'][shortcut])
        else:
            pass

    def get_ui_root(self):
        found = False
        obj = self
        while not found:
            if obj.__class__.__name__ == 'ISCSIRoot':
                break
            obj = obj.parent
        return obj


class UIGroup(UICommon):

    def __init__(self, name, parent=None, shell=None):
        UICommon.__init__(self, name, parent, shell)
        self.http_mode = self.parent.http_mode

    def reset(self):
        children = set(self.children)  # set of child objects
        for child in children:
            self.remove_child(child)


class UINode(UIGroup):

    display_attributes = None

    def __init__(self, name, parent):
        UIGroup.__init__(self, name, parent)
        self.http_mode = self.parent.http_mode

    def ui_command_info(self):
        """
        Show the attributes of the current object.
        """

        text = self.get_info()

        console_message(text)

    def get_info(self):
        """
        extract the relevant display fields from the object and format
        ready for printing
        :return: (str) object meta data based on object's display_attributes
                 list
        """

        display_text = ''

        if not self.display_attributes:
            return "'info' not available for this item"

        field_list = self.display_attributes
        max_field_size = len(max(field_list, key=len))
        for k in field_list:
            attr_label = k.replace('_', ' ').title()

            attr_value = getattr(self, k)

            if isinstance(attr_value, dict):

                if attr_value:
                    display_text += "{}\n".format(attr_label)
                    max_dict_field = len(max(attr_value.keys(), key=len))
                    for dict_key in sorted(attr_value):

                        if isinstance(attr_value[dict_key], dict):
                            inner_dict = attr_value[dict_key]
                            display_value = ", ".join(["=".join(
                                [key, str(val)]) for key, val in inner_dict.items()])
                            display_text += ("- {:<{}} .. {}\n".format(dict_key,
                                                                       max_dict_field,
                                                                       display_value))

                        else:
                            display_text += ("- {} .. {}\n".format(dict_key,
                                                                   attr_value[dict_key]))

                    continue
                else:
                    attr_value = 'UNDEFINED\n'

            if isinstance(attr_value, list):
                item_1 = True
                attr_string = ''
                for item in attr_value:
                    if item_1:
                        attr_string = "{}\n".format(str(item))
                        item_1 = False
                    else:
                        attr_string += "{}{}\n".format(" " * (max_field_size + 4),
                                                       str(item))

                attr_value = attr_string[:-1]

            display_text += ("{:<{}} .. {}\n".format(attr_label,
                                                     max_field_size,
                                                     attr_value))

        return display_text


class UIRoot(UICommon):
    """
    The gwcli hierarchy root node.
    """

    def __init__(self, shell, as_root=False):
        UICommon.__init__(self, '/', shell=shell)
        self.as_root = as_root



import rados
import rbd
import re

from time import sleep

from rtslib_fb import UserBackedStorageObject, root
from rtslib_fb.utils import RTSLibError

import ceph_iscsi_config.settings as settings

from ceph_iscsi_config.gateway_setting import TCMU_SETTINGS
from ceph_iscsi_config.backstore import USER_RBD
from ceph_iscsi_config.utils import (convert_2_bytes, gen_control_string,
                                     valid_size, get_pool_id, ip_addresses,
                                     get_pools, get_rbd_size, this_host,
                                     human_size, CephiSCSIError)
from ceph_iscsi_config.gateway_object import GWObject
from ceph_iscsi_config.target import GWTarget
from ceph_iscsi_config.client import GWClient, CHAP
from ceph_iscsi_config.group import Group
from ceph_iscsi_config.backstore import lookup_storage_object

__author__ = 'pcuzner@redhat.com'


class RBDDev(object):

    unsupported_features_list = {
        USER_RBD: []
    }

    default_features_list = {
        USER_RBD: [
            'RBD_FEATURE_LAYERING',
            'RBD_FEATURE_EXCLUSIVE_LOCK',
            'RBD_FEATURE_OBJECT_MAP',
            'RBD_FEATURE_FAST_DIFF',
            'RBD_FEATURE_DEEP_FLATTEN'
        ]
    }

    required_features_list = {
        USER_RBD: [
            'RBD_FEATURE_EXCLUSIVE_LOCK'
        ]
    }

    def __init__(self, image, size, backstore, pool=None):
        self.image = image
        self.size_bytes = convert_2_bytes(size)
        self.backstore = backstore
        if pool is None:
            pool = settings.config.pool
        self.pool = pool
        self.pool_id = get_pool_id(pool_name=self.pool)
        self.error = False
        self.error_msg = ''
        self.changed = False

    def create(self):
        """
        Create an rbd image compatible with exporting through LIO to multiple
        clients
        :return: status code and msg
        """

        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:
            with cluster.open_ioctx(self.pool) as ioctx:
                rbd_inst = rbd.RBD()
                try:
                    rbd_inst.create(ioctx,
                                    self.image,
                                    self.size_bytes,
                                    features=RBDDev.default_features(self.backstore),
                                    old_format=False)

                except (rbd.ImageExists, rbd.InvalidArgument) as err:
                    self.error = True
                    self.error_msg = ("Failed to create rbd image {} in "
                                      "pool {} : {}".format(self.image,
                                                            self.pool,
                                                            err))

    def delete(self):
        """
        Delete the current rbd image
        :return: nothing, but the objects error attribute is set if there
        are problems
        """

        rbd_deleted = False
        extra_error_info = ''

        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:
            with cluster.open_ioctx(self.pool) as ioctx:
                rbd_inst = rbd.RBD()

                ctr = 0
                while ctr < settings.config.time_out:

                    try:
                        rbd_inst.remove(ioctx, self.image)
                    except rbd.ImageNotFound:
                        rbd_deleted = True
                        break
                    except rbd.ImageBusy:
                        # catch and ignore the busy state - rbd probably still mapped on
                        # another gateway, so we keep trying
                        pass
                    except rbd.ImageHasSnapshots:
                        extra_error_info = " - Image has snapshots"
                        break
                    else:
                        rbd_deleted = True
                        break

                    sleep(settings.config.loop_delay)
                    ctr += settings.config.loop_delay

                if rbd_deleted:
                    return
                else:
                    self.error = True
                    self.error_msg = ("Unable to delete the underlying rbd "
                                      "image {}".format(self.image))
                    if extra_error_info:
                        self.error_msg += extra_error_info

    def rbd_size(self):
        """
        Confirm that the existing rbd image size, matches the requirement
        passed in the request - if the required size is > than
        current, resize the rbd image to match
        :return: boolean value reflecting whether the rbd image was resized
        """

        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:
            with cluster.open_ioctx(self.pool) as ioctx:
                with rbd.Image(ioctx, self.image) as rbd_image:

                    # get the current size in bytes
                    current_bytes = rbd_image.size()

                    if self.size_bytes > current_bytes:

                        # resize method, doesn't document potential exceptions
                        # so using a generic catch all (Yuk!)
                        try:
                            rbd_image.resize(self.size_bytes)
                        except Exception:
                            self.error = True
                            self.error_msg = ("rbd image resize failed for "
                                              "{}".format(self.image))
                        else:
                            self.changed = True

    def _get_size_bytes(self):
        """
        Return the current size of the rbd image
        :return: (int) rbd image size in bytes
        """

        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:
            with cluster.open_ioctx(self.pool) as ioctx:
                with rbd.Image(ioctx, self.image) as rbd_image:
                    image_size = rbd_image.size()

        return image_size

    @staticmethod
    def rbd_list(conf=None, pool=None):
        """
        return a list of rbd images in a given pool
        :param pool: pool name (str) to return a list of rbd image names for
        :return: list of rbd image names (list)
        """

        if conf is None:
            conf = settings.config.cephconf
        if pool is None:
            pool = settings.config.pool

        with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
            with cluster.open_ioctx(pool) as ioctx:
                rbd_inst = rbd.RBD()
                rbd_names = rbd_inst.list(ioctx)
        return rbd_names

    @staticmethod
    def rbd_lock_cleanup(logger, local_ips, rbd_image):
        """
        cleanup locks left if this node crashed and was not able to release them
        :param logger: logger object to print to
        :param local_ips: list of local ip addresses.
        :rbd_image: rbd image to clean up locking for
        :raise CephiSCSIError.
        """

        lock_info = rbd_image.list_lockers()
        if not lock_info:
            return

        lockers = lock_info.get("lockers")
        for holder in lockers:
            for ip in local_ips:
                if ip in holder[2]:
                    logger.info("Cleaning up stale local lock for {} {}".format(
                                holder[0], holder[1]))
                    try:
                        rbd_image.break_lock(holder[0], holder[1])
                    except Exception as err:
                        raise CephiSCSIError("Could not break lock for {}. "
                                             "Error {}".format(rbd_image, err))

    def _valid_rbd(self):

        valid_state = True
        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:
            ioctx = cluster.open_ioctx(self.pool)
            with rbd.Image(ioctx, self.image) as rbd_image:

                if rbd_image.features() & RBDDev.required_features(self.backstore) != \
                        RBDDev.required_features(self.backstore):
                    valid_state = False

        return valid_state

    @classmethod
    def unsupported_features(cls, backstore):
        """
        Return an int representing the unsupported features for LIO export
        :return: int
        """
        # build the required feature settings into an int
        feature_int = 0
        for feature in RBDDev.unsupported_features_list[backstore]:
            feature_int += getattr(rbd, feature)

        return feature_int

    @classmethod
    def default_features(cls, backstore):
        """
        Return an int representing the default features for image creation
        :return: int
        """
        # build the required feature settings into an int
        feature_int = 0
        for feature in RBDDev.default_features_list[backstore]:
            feature_int += getattr(rbd, feature)

        return feature_int

    @classmethod
    def required_features(cls, backstore):
        """
        Return an int representing the required features for LIO export
        :return: int
        """
        # build the required feature settings into an int
        feature_int = 0
        for feature in RBDDev.required_features_list[backstore]:
            feature_int += getattr(rbd, feature)

        return feature_int

    current_size = property(_get_size_bytes,
                            doc="return the current size of the rbd(bytes)")

    valid = property(_valid_rbd,
                     doc="check the rbd is valid for export through LIO"
                         " (boolean)")


class LUN(GWObject):
    BACKSTORES = [
        USER_RBD
    ]

    DEFAULT_BACKSTORE = USER_RBD

    SETTINGS = {
        USER_RBD: TCMU_SETTINGS
    }

    def __init__(self, logger, pool, image, size, allocating_host,
                 backstore, backstore_object_name):
        self.logger = logger
        self.image = image
        self.pool = pool
        self.pool_id = 0
        self.size_bytes = convert_2_bytes(size)
        self.config_key = '{}/{}'.format(self.pool, self.image)

        self.allocating_host = allocating_host
        self.backstore = backstore
        self.backstore_object_name = backstore_object_name

        self.error = False
        self.error_msg = ''
        self.num_changes = 0

        try:
            super(LUN, self).__init__('disks', self.config_key, logger,
                                      LUN.SETTINGS[self.backstore])
        except CephiSCSIError as err:
            self.error = True
            self.error_msg = err

        self._validate_request()

    def _validate_request(self):

        if not rados_pool(pool=self.pool):
            # Could create the pool, but a fat finger moment in the config
            # file would mean rbd images get created and mapped, and then need
            # correcting. Better to exit if the pool doesn't exist
            self.error = True
            self.error_msg = ("Pool '{}' does not exist. Unable to "
                              "continue".format(self.pool))

    def remove_lun(self, preserve_image):
        local_gw = this_host()
        self.logger.info("LUN deletion request received, rbd removal to be "
                         "performed by {}".format(self.allocating_host))

        # First ensure the LUN is not in use
        for target_iqn, target in self.config.config['targets'].items():
            if self.config_key in target['disks']:
                self.error = True
                self.error_msg = ("Unable to delete {} - allocated to "
                                  "{}".format(self.config_key,
                                              target_iqn))

                self.logger.warning(self.error_msg)
                return

        # Check that the LUN is in LIO - if not there is nothing to do for
        # this request
        lun = self.lio_stg_object()
        if lun:
            # Now we know the request is for a LUN in LIO, and it's not masked
            # to a client
            self.remove_dev_from_lio()
            if self.error:
                return

        rbd_image = RBDDev(self.image, '0G', self.backstore, self.pool)

        if local_gw == self.allocating_host:
            # by using the allocating host we ensure the delete is not
            # issue by several hosts when initiated through ansible
            if not preserve_image:
                rbd_image.delete()
                if rbd_image.error:
                    self.error = True
                    self.error_msg = rbd_image.error_msg
                    return

            # remove the definition from the config object
            self.config.del_item('disks', self.config_key)

            self.config.commit()

    def unmap_lun(self, target_iqn):
        local_gw = this_host()
        self.logger.info("LUN unmap request received, config commit to be "
                         "performed by {}".format(self.allocating_host))

        target_config = self.config.config['targets'][target_iqn]

        # First ensure the LUN is not in use
        clients = target_config['clients']
        for client_iqn in clients:
            client_luns = clients[client_iqn]['luns'].keys()
            if self.config_key in client_luns:
                self.error = True
                self.error_msg = ("Unable to delete {} - allocated to {}"
                                  .format(self.config_key, client_iqn))
                self.logger.warning(self.error_msg)
                return

        # Check that the LUN is in LIO - if not there is nothing to do for
        # this request
        lun = self.lio_stg_object()
        if not lun:
            return

        # Now we know the request is for a LUN in LIO, and it's not masked
        # to a client
        self.remove_dev_from_lio()
        if self.error:
            return

        if local_gw == self.allocating_host:
            # by using the allocating host we ensure the delete is not
            # issue by several hosts when initiated through ansible

            target_config['disks'].pop(self.config_key)
            self.config.update_item("targets", target_iqn, target_config)

            # determine which host was the path owner
            disk_owner = self.config.config['disks'][self.config_key]['owner']

            # update the active_luns count for gateway that owned this
            # lun
            gw_metadata = self.config.config['gateways'][disk_owner]
            if gw_metadata['active_luns'] > 0:
                gw_metadata['active_luns'] -= 1

                self.config.update_item('gateways',
                                        disk_owner,
                                        gw_metadata)

            disk_metadata = self.config.config['disks'][self.config_key]
            if 'owner' in disk_metadata:
                del disk_metadata['owner']
                self.logger.debug("{} owner deleted".format(self.config_key))
            self.config.update_item("disks", self.config_key, disk_metadata)

            self.config.commit()

    def _get_next_lun_id(self, target_disks):
        lun_ids_in_use = [t['lun_id'] for t in target_disks.values()]
        lun_id_candidate = 0
        while lun_id_candidate in lun_ids_in_use:
            lun_id_candidate += 1
        return lun_id_candidate

    def map_lun(self, gateway, owner, disk, lun_id=None):
        target_config = self.config.config['targets'][gateway.iqn]
        disk_metadata = self.config.config['disks'][disk]
        disk_metadata['owner'] = owner
        self.config.update_item("disks", disk, disk_metadata)

        target_disk_config = target_config['disks'].get(disk)
        if not target_disk_config:
            if lun_id is None:
                lun_id = self._get_next_lun_id(target_config['disks'])
            target_config['disks'][disk] = {
                'lun_id': lun_id
            }
        self.config.update_item("targets", gateway.iqn, target_config)

        gateway_dict = self.config.config['gateways'][owner]
        gateway_dict['active_luns'] += 1
        self.config.update_item('gateways', owner, gateway_dict)

        so = self.allocate()
        if self.error:
            raise CephiSCSIError(self.error_msg)

        gateway.map_lun(self.config, so, target_config['disks'][disk])
        if gateway.error:
            raise CephiSCSIError(gateway.error_msg)

    def manage(self, desired_state):

        self.logger.debug("LUN.manage request for {}, desired state "
                          "{}".format(self.image, desired_state))

        if desired_state == 'present':

            self.allocate()

        elif desired_state == 'absent':

            self.remove_lun()

    def deactivate(self):
        so = self.lio_stg_object()
        if not so:
            # Could be due to a restart after failure. Just log and ignore.
            self.logger.warning("LUN {} already deactivated".format(self.image))
            return

        for alun in so.attached_luns:
            for mlun in alun.mapped_luns:
                node_acl = mlun.parent_nodeacl

                if node_acl.session and \
                        node_acl.session.get('state', '').upper() == 'LOGGED_IN':
                    raise CephiSCSIError("LUN {} in-use".format(self.image))

        self.remove_dev_from_lio()
        if self.error:
            raise CephiSCSIError("LUN deactivate failure - {}".format(self.error_msg))

    def activate(self):
        disk = self.config.config['disks'].get(self.config_key, None)
        if not disk:
            raise CephiSCSIError("Image {} not found.".format(self.image))

        wwn = disk.get('wwn', None)
        if not wwn:
            raise CephiSCSIError("LUN {} missing wwn".format(self.image))

        # re-add backend storage object
        so = self.lio_stg_object()
        if not so:
            so = self.add_dev_to_lio(wwn)
            if self.error:
                raise CephiSCSIError("LUN activate failure - {}".format(self.error_msg))

        # re-add LUN to target
        local_gw = this_host()
        targets_items = [item for item in self.config.config['targets'].items()
                         if self.config_key in item[1]['disks'] and local_gw in item[1]['portals']]
        for target_iqn, target in targets_items:
            ip_list = target['ip_list']

            # Add the mapping for the lun to ensure the block device is
            # present on all TPG's
            gateway = GWTarget(self.logger, target_iqn, ip_list)
            gateway.map_lun(self.config, so, target['disks'][self.config_key])
            if gateway.error:
                raise CephiSCSIError("LUN mapping failed - {}".format(gateway.error_msg))

            # re-map LUN to hosts
            client_err = ''
            for client_iqn in target['clients']:
                client_metadata = target['clients'][client_iqn]
                if client_metadata.get('group_name', ''):
                    continue

                image_list = list(client_metadata['luns'].keys())
                if self.config_key not in image_list:
                    continue

                client_auth_config = client_metadata['auth']

                client_chap = CHAP(client_auth_config['username'],
                                   client_auth_config['password'],
                                   client_auth_config['password_encryption_enabled'])
                if client_chap.error:
                    raise CephiSCSIError("Password decode issue : "
                                         "{}".format(client_chap.error_msg))

                client_chap_mutual = CHAP(client_auth_config['mutual_username'],
                                          client_auth_config['mutual_password'],
                                          client_auth_config['mutual_password_encryption_enabled'])
                if client_chap_mutual.error:
                    raise CephiSCSIError("Password decode issue : "
                                         "{}".format(client_chap_mutual.error_msg))

                client = GWClient(self.logger, client_iqn, image_list, client_chap.user,
                                  client_chap.password, client_chap_mutual.user,
                                  client_chap_mutual.password, target_iqn)
                client.manage('present')
                if client.error:
                    client_err = "LUN mapping failed {} - {}".format(client_iqn,
                                                                     client.error_msg)

            # re-map LUN to host groups
            for group_name in target['groups']:
                host_group = target['groups'][group_name]
                members = host_group.get('members')
                disks = host_group.get('disks').keys()
                if self.config_key not in disks:
                    continue

                group = Group(self.logger, target_iqn, group_name, members, disks)
                group.apply()
                if group.error:
                    client_err = "LUN mapping failed {} - {}".format(group_name,
                                                                     group.error_msg)

            if client_err:
                raise CephiSCSIError(client_err)

    def allocate(self, keep_dev_in_lio=True, in_wwn=None):
        """
        Create image and add to LIO and config.

        :param keep_dev_in_lio: (bool) false if the LIO so should be removed
                                 after allocating the wwn.
        :return: LIO storage object if successful and keep_dev_in_lio=True
                 else None.
        """
        self.logger.debug("LUN.allocate starting, listing rbd devices")
        disk_list = RBDDev.rbd_list(pool=self.pool)
        self.logger.debug("rados pool '{}' contains the following - "
                          "{}".format(self.pool, disk_list))

        local_gw = this_host()
        self.logger.debug("Hostname Check - this host is {}, target host for "
                          "allocations is {}".format(local_gw,
                                                     self.allocating_host))

        rbd_image = RBDDev(self.image, self.size_bytes, self.backstore, self.pool)
        self.pool_id = rbd_image.pool_id

        # if the image required isn't defined, create it!
        if self.image not in disk_list:
            # create the requested disk if this is the 'owning' host
            if local_gw == self.allocating_host:

                rbd_image.create()

                if not rbd_image.error:
                    self.config.add_item('disks', self.config_key)
                    self.logger.info("(LUN.allocate) created {}/{} "
                                     "successfully".format(self.pool,
                                                           self.image))
                    self.num_changes += 1
                else:
                    self.error = True
                    self.error_msg = rbd_image.error_msg
                    return None

            else:
                # the image isn't there, and this isn't the 'owning' host
                # so wait until the disk arrives
                waiting = 0
                while self.image not in disk_list:
                    sleep(settings.config.loop_delay)
                    disk_list = RBDDev.rbd_list(pool=self.pool)
                    waiting += settings.config.loop_delay
                    if waiting >= settings.config.time_out:
                        self.error = True
                        self.error_msg = ("(LUN.allocate) timed out waiting "
                                          "for rbd to show up")
                        return None
        else:
            # requested image is already defined to ceph

            if rbd_image.valid:
                # rbd image is OK to use, so ensure it's in the config
                # object
                if self.config_key not in self.config.config['disks']:
                    self.config.add_item('disks', self.config_key)

            else:
                # rbd image is not valid for export, so abort
                self.error = True
                features = ','.join(RBDDev.unsupported_features_list[self.backstore])
                self.error_msg = ("(LUN.allocate) rbd '{}' is not compatible "
                                  "with LIO\nImage features {} are not"
                                  " supported".format(self.image, features))
                self.logger.error(self.error_msg)
                return None

        self.logger.debug("Check the rbd image size matches the request")

        # if updates_made is not set, the disk pre-exists so on the owning
        # host see if it needs to be resized
        if self.num_changes == 0 and local_gw == self.allocating_host:

            # check the size, and update if needed
            rbd_image.rbd_size()
            if rbd_image.error:
                self.logger.critical(rbd_image.error_msg)
                self.error = True
                self.error_msg = rbd_image.error_msg
                return None

            if rbd_image.changed:
                self.logger.info("rbd image {} resized "
                                 "to {}".format(self.config_key,
                                                self.size_bytes))
                self.num_changes += 1
            else:
                self.logger.debug("rbd image {} size matches the configuration"
                                  " file request".format(self.config_key))

        self.logger.debug("Begin processing LIO mapping")

        # now see if we need to add this rbd image to LIO
        so = self.lio_stg_object()
        if not so:

            # this image has not been defined to this hosts LIO, so check the
            # config for the details and if it's  missing define the
            # wwn/alua_state and update the config
            if local_gw == self.allocating_host:
                # first check to see if the device needs adding
                try:
                    wwn = self.config.config['disks'][self.config_key]['wwn']
                except KeyError:
                    wwn = ''

                if wwn == '' or in_wwn is not None:
                    # disk hasn't been defined to LIO yet, it' not been defined
                    # to the config yet and this is the allocating host
                    so = self.add_dev_to_lio(in_wwn)
                    if self.error:
                        return None

                    # lun is now in LIO, time for some housekeeping :P
                    wwn = so._get_wwn()

                    if not keep_dev_in_lio:
                        self.remove_dev_from_lio()
                        if self.error:
                            return None

                    disk_attr = {"wwn": wwn,
                                 "image": self.image,
                                 "pool": self.pool,
                                 "allocating_host": self.allocating_host,
                                 "pool_id": rbd_image.pool_id,
                                 "controls": self.controls,
                                 "backstore": self.backstore,
                                 "backstore_object_name": self.backstore_object_name}

                    self.config.update_item('disks',
                                            self.config_key,
                                            disk_attr)

                    self.logger.debug("(LUN.allocate) registered '{}' with "
                                      "wwn '{}' with the config "
                                      "object".format(self.image,
                                                      wwn))
                    self.logger.info("(LUN.allocate) added '{}/{}' to LIO and"
                                     " config object".format(self.pool,
                                                             self.image))

                else:
                    # config object already had wwn for this rbd image
                    so = self.add_dev_to_lio(wwn)
                    if self.error:
                        return None

                    self.update_controls()
                    self.logger.debug("(LUN.allocate) registered '{}' to LIO "
                                      "with wwn '{}' from the config "
                                      "object".format(self.image,
                                                      wwn))

                self.num_changes += 1

            else:
                # lun is not already in LIO, but this is not the owning node
                # that defines the wwn we need the wwn from the config
                # (placed by the allocating host), so we wait!
                waiting = 0
                while waiting < settings.config.time_out:
                    self.config.refresh()
                    if self.config_key in self.config.config['disks']:
                        if 'wwn' in self.config.config['disks'][self.config_key]:
                            if self.config.config['disks'][self.config_key]['wwn']:
                                wwn = self.config.config['disks'][self.config_key]['wwn']
                                break
                    sleep(settings.config.loop_delay)
                    waiting += settings.config.loop_delay
                    self.logger.debug("(LUN.allocate) waiting for config object"
                                      " to show {} with it's wwn".format(self.image))

                if waiting >= settings.config.time_out:
                    self.error = True
                    self.error_msg = ("(LUN.allocate) waited too long for the "
                                      "wwn information on image {} to "
                                      "arrive".format(self.image))
                    return None

                # At this point we have a wwn from the config for this rbd
                # image, so just add to LIO
                so = self.add_dev_to_lio(wwn)
                if self.error:
                    return None

                self.logger.info("(LUN.allocate) added {} to LIO using wwn "
                                 "'{}' defined by {}".format(self.image,
                                                             wwn,
                                                             self.allocating_host))

                self.num_changes += 1

        else:
            # lun exists in LIO, check the size is correct
            if not self.lio_size_ok(rbd_image, so):
                self.error = True
                self.error_msg = "Unable to sync the rbd device size with LIO"
                self.logger.critical(self.error_msg)
                return None

        self.logger.debug("config meta data for this disk is "
                          "{}".format(self.config.config['disks'][self.config_key]))

        # the owning host for an image is the only host that commits to the
        # config
        if local_gw == self.allocating_host and self.config.changed:

            self.logger.debug("(LUN.allocate) Committing change(s) to the "
                              "config object in pool {}".format(self.pool))
            self.config.commit()
            self.error = self.config.error
            self.error_msg = self.config.error_msg
            if self.error:
                return None

        return so

    def lio_size_ok(self, rbd_object, stg_object):
        """
        Check that the SO in LIO matches the current size of the rbd. if the
        size requested < current size, just return. Downsizing an rbd is not
        supported by this code and problematic for client filesystems anyway!
        :return boolean indicating whether the size matches
        """

        tmr = 0
        size_ok = False
        rbd_size_ok = False
        # dm_path_found = False

        # We have to wait for the rbd size to match, since the rbd could have
        # been resized on another gateway host
        while tmr < settings.config.time_out:
            if self.size_bytes <= rbd_object.current_size:
                rbd_size_ok = True
                break
            sleep(settings.config.loop_delay)
            tmr += settings.config.loop_delay

        # we have the right size for the rbd - check that LIO dev size matches
        if rbd_size_ok:

            # If the LIO size is not right, poke it with the new value
            if stg_object.size < self.size_bytes:
                self.logger.info("Resizing {} in LIO "
                                 "to {}".format(self.config_key,
                                                self.size_bytes))

                stg_object.set_attribute("dev_size", self.size_bytes)

                size_ok = stg_object.size == self.size_bytes

            else:
                size_ok = True

        return size_ok

    def lio_stg_object(self):
        try:
            return lookup_storage_object(self.backstore_object_name, self.backstore)
        except RTSLibError as err:
            self.logger.debug("lio stg lookup failed {}".format(err))
            return None

    def add_dev_to_lio(self, in_wwn=None):
        """
        Add an rbd device to the LIO configuration
        :param in_wwn: optional wwn identifying the rbd image to clients
        (must match across gateways)
        :return: LIO LUN object
        """
        self.logger.info("(LUN.add_dev_to_lio) Adding image "
                         "'{}' to LIO backstore {}".format(self.config_key, self.backstore))

        new_lun = None
        if self.backstore == USER_RBD:
            new_lun = self._add_dev_to_lio_user_rbd(in_wwn)

        else:
            raise CephiSCSIError("Error adding device to lio - "
                                 "Unsupported backstore {}".format(self.backstore))

        if new_lun:
            self.logger.info("(LUN.add_dev_to_lio) Successfully added {}"
                             " to LIO".format(self.config_key))

        return new_lun

    def _add_dev_to_lio_user_rbd(self, in_wwn=None):
        """
        Add an rbd device to the LIO configuration (`USER_RBD`)
        :param in_wwn: optional wwn identifying the rbd image to clients
        (must match across gateways)
        :return: LIO LUN object
        """
        # extract control parameter overrides (if any) or use default
        controls = {}
        for k in ['max_data_area_mb', 'hw_max_sectors']:
            controls[k] = getattr(self, k)

        control_string = gen_control_string(controls)
        if control_string:
            self.logger.debug("control=\"{}\"".format(control_string))

        new_lun = None
        try:
            # config string = rbd identifier / config_key (pool/image) /
            # optional osd timeout
            cfgstring = "rbd/{}/{};osd_op_timeout={}".format(self.pool,
                                                             self.image,
                                                             self.osd_op_timeout)
            if (settings.config.cephconf != '/etc/ceph/ceph.conf'):
                cfgstring += ";conf={}".format(settings.config.cephconf)

            if (settings.config.cluster_client_name != 'client.admin'):
                client_id = settings.config.cluster_client_name.split('.', 1)[1]
                cfgstring += ";id={}".format(client_id)

            new_lun = UserBackedStorageObject(name=self.backstore_object_name,
                                              config=cfgstring,
                                              size=self.size_bytes,
                                              wwn=in_wwn, control=control_string)
        except (RTSLibError, IOError) as err:
            self.error = True
            self.error_msg = ("failed to add {} to LIO - "
                              "error({})".format(self.config_key,
                                                 str(err)))
            self.logger.error(self.error_msg)
            return None

        try:
            new_lun.set_attribute("cmd_time_out", 0)
            new_lun.set_attribute("qfull_time_out", self.qfull_timeout)
        except RTSLibError as err:
            self.error = True
            self.error_msg = ("Could not set LIO device attribute "
                              "cmd_time_out/qfull_time_out for device: {}. "
                              "Kernel not supported. - "
                              "error({})".format(self.config_key, str(err)))
            self.logger.error(self.error_msg)
            new_lun.delete()
            return None

        return new_lun

    def remove_dev_from_lio(self):
        lio_root = root.RTSRoot()

        # remove the device from all tpgs
        for t in lio_root.tpgs:
            for lun in t.luns:
                if lun.storage_object.name == self.backstore_object_name:
                    try:
                        lun.delete()
                    except Exception as e:
                        self.error = True
                        self.error_msg = ("Delete from LIO/TPG failed - "
                                          "{}".format(e))
                        return
                    else:
                        break       # continue to the next tpg

        so = self.lio_stg_object()
        if not so:
            self.error = True
            self.error_msg = ("Removal failed. Could not find LIO object.")
            return

        try:
            so.delete()
        except Exception as err:
            self.error = True
            self.error_msg = ("Delete from LIO/backstores failed - "
                              "{}".format(err))
            return

    @staticmethod
    def valid_disk(ceph_iscsi_config, logger, **kwargs):
        """
        determine whether the given image info is valid for a disk operation

        :param ceph_iscsi_config: Config object
        :param logger: logger object
        :param image_id: (str) <pool>.<image> format
        :return: (str) either 'ok' or an error description
        """

        # create can also pass optional controls dict
        mode_vars = {"create": ['pool', 'image', 'size', 'count'],
                     "resize": ['pool', 'image', 'size'],
                     "reconfigure": ['pool', 'image', 'controls'],
                     "delete": ['pool', 'image']}

        if 'mode' in kwargs.keys():
            mode = kwargs['mode']
        else:
            mode = None

        backstore = kwargs['backstore']
        if backstore not in LUN.BACKSTORES:
            return "Invalid '{}' backstore - Supported backstores: " \
                   "{}".format(backstore, ','.join(LUN.BACKSTORES))

        if mode in mode_vars:
            if not all(x in kwargs for x in mode_vars[mode]):
                return ("{} request must contain the following "
                        "variables: ".format(mode,
                                             ','.join(mode_vars[mode])))
        else:
            return "disk operation mode '{}' is invalid".format(mode)

        config = ceph_iscsi_config.config

        disk_key = "{}/{}".format(kwargs['pool'], kwargs['image'])

        if mode in ['create', 'resize']:

            if kwargs['pool'] not in get_pools():
                return "pool name is invalid"

        if mode == 'create':
            if kwargs['size'] and not valid_size(kwargs['size']):
                return "Size is invalid"

            if len(config['disks']) >= 256:
                return "Disk limit of 256 reached."

            disk_regex = re.compile(r"^[a-zA-Z0-9\-_\.]+$")
            if not disk_regex.search(kwargs['pool']):
                return "Invalid pool name (use alphanumeric, '_', '.', or '-' characters)"
            if not disk_regex.search(kwargs['image']):
                return "Invalid image name (use alphanumeric, '_', '.', or '-' characters)"

            if kwargs['wwn'] is not None:
                for disk_id, disk_config in config['disks'].items():
                    if disk_config['wwn'] == kwargs['wwn']:
                        return "WWN {} is already in use by {}".format(kwargs['wwn'], disk_id)

            if kwargs['count'].isdigit():
                if not 1 <= int(kwargs['count']) <= 10:
                    return "invalid count specified, must be an integer (1-10)"
                if int(kwargs['count']) > 1 and kwargs['wwn'] is not None:
                    return "WWN cannot be specified when count > 1"
            else:
                return "invalid count specified, must be an integer (1-10)"

            if kwargs['count'] == '1':
                new_disks = {disk_key}
            else:
                limit = int(kwargs['count']) + 1
                new_disks = set(['{}{}'.format(disk_key, ctr)
                                 for ctr in range(1, limit)])

            if any(new_disk in config['disks'] for new_disk in new_disks):
                return ("at least one rbd image(s) with that name/prefix is "
                        "already defined")

        if mode in ["resize", "delete", "reconfigure"]:
            # disk must exist in the config
            if disk_key not in config['disks']:
                return ("rbd {}/{} is not defined to the "
                        "configuration".format(kwargs['pool'],
                                               kwargs['image']))

        if mode == 'resize':

            if not valid_size(kwargs['size']):
                return "Size is invalid"

            size = kwargs['size'].upper()
            current_size = get_rbd_size(kwargs['pool'], kwargs['image'])
            if convert_2_bytes(size) <= current_size:
                return ("resize value must be larger than the "
                        "current size ({}/{})".format(human_size(current_size),
                                                      current_size))

        if mode in ['create', 'reconfigure']:

            try:
                settings.Settings.normalize_controls(kwargs['controls'],
                                                     LUN.SETTINGS[backstore])
            except ValueError as err:
                return(err)

        if mode == 'delete':

            # disk must *not* be allocated to a client in the config
            mapped_list = []
            allocation_list = []
            for target_iqn, target in config['targets'].items():
                if disk_key in target['disks']:
                    mapped_list.append(target_iqn)
                for client_iqn in target['clients']:
                    client_metadata = target['clients'][client_iqn]
                    if disk_key in client_metadata['luns']:
                        allocation_list.append(client_iqn)

            if allocation_list:
                return ("Unable to delete {}. Allocated "
                        "to: {}".format(disk_key,
                                        ','.join(allocation_list)))

            if mapped_list:
                return ("Unable to delete {}. Mapped "
                        "to: {}".format(disk_key,
                                        ','.join(mapped_list)))

        return 'ok'

    @staticmethod
    def get_owner(gateways, portals):
        """
        Determine the gateway in the configuration with the lowest number of
        active LUNs. This gateway is then selected as the owner for the
        primary path of the current LUN being processed
        :param gateways: gateway dict returned from the RADOS configuration
               object
        :param portals: portal dict returned from the RADOS configuration
               object
        :return: specific gateway hostname (str) that should provide the
               active path for the next LUN
        """

        return sorted(portals.keys(),
                      key=lambda x: (gateways[x]['active_luns']))[0]

    @staticmethod
    def _backstore_object_name_exists(disks_config, backstore_object_name_exists):
        return len([disk for _, disk in disks_config.items()
                    if disk['backstore_object_name'] == backstore_object_name_exists]) > 0

    @staticmethod
    def get_backstore_object_name(pool, image, disks_config):
        """
        Determine the backstore storage object name based on the pool name,
        image name, and existing storage object names to avoid conflicts.

        Example of how name conflict resolution will work:
          - Add disk `a.b/c` will create the storage object `a.b.c`
          - Add disk `a/b.c` will create the storage object `a.b.c.1`

        :param pool: pool name
        :param image: image name
        :param disks_config: disks configuration from `gateway.conf`
        :return: the backstore storage object name to be used
        """
        base_name = '{}.{}'.format(pool, image)
        candidate = base_name
        counter = 0
        while LUN._backstore_object_name_exists(disks_config, candidate):
            counter += 1
            candidate = '{}.{}'.format(base_name, counter)
        return candidate

    @staticmethod
    def find_first_mapped_target(config, disk):
        for target, target_config in config.config['targets'].items():
            if disk in target_config['disks']:
                return target

        return None

    @staticmethod
    def reassign_owners(logger, config):
        """
        Reassign disks across gateways after a gw deletion.
        :param logger: logger object to print to
        :param config: configuration dict from the rados pool
        :raises CephiSCSIError.
        """

        updated = False
        gateways = config.config['gateways']

        for disk, disk_config in config.config['disks'].items():
            owner = disk_config.get('owner', None)
            if owner is None:
                continue

            gw = gateways.get(owner, None)
            if gw is None:
                target = LUN.find_first_mapped_target(config, disk)

                if not gateways or target is None:
                    disk_config.pop('owner')
                else:
                    target_config = config.config['targets'][target]
                    new_owner = LUN.get_owner(gateways,
                                              target_config['portals'])

                    logger.info("Changing {}'s owner from {} to {}".
                                format(disk, owner, new_owner))
                    disk_config['owner'] = new_owner

                    gw_config = config.config['gateways'][new_owner]
                    active_cnt = gw_config['active_luns']
                    gw_config['active_luns'] = active_cnt + 1
                    config.update_item("gateways", new_owner, gw_config)

                config.update_item("disks", disk, disk_config)
                updated = True

        if updated:
            config.commit("retain")
            if config.error:
                raise CephiSCSIError("Could not update LUN owners: {}".
                                     format(config.error_msg))

    @staticmethod
    def define_luns(logger, config, target):
        """
        define the disks in the config to LIO and map to a LUN
        :param logger: logger object to print to
        :param config: configuration dict from the rados pool
        :param target: (object) gateway object - used for mapping
        :raises CephiSCSIError.
        """

        ips = ip_addresses()
        local_gw = this_host()

        target_disks = config.config["targets"][target.iqn]['disks']
        if not target_disks:
            logger.info("No LUNs to export")
            return

        disks = {}
        for disk in target_disks.keys():
            disks[disk] = config.config['disks'][disk]

        # sort the disks dict keys, so the disks are registered in a specific
        # sequence
        srtd_disks = sorted(disks)
        pools = {disks[disk_key]['pool'] for disk_key in srtd_disks}

        ips = ip_addresses()

        with rados.Rados(conffile=settings.config.cephconf,
                         name=settings.config.cluster_client_name) as cluster:

            for pool in pools:

                logger.debug("Processing rbd's in '{}' pool".format(pool))

                with cluster.open_ioctx(pool) as ioctx:

                    pool_disks = [disk_key for disk_key in srtd_disks
                                  if disk_key.startswith(pool + '/')]
                    for disk_key in pool_disks:

                        pool, image_name = disk_key.split('/')
                        with rbd.Image(ioctx, image_name) as rbd_image:

                            disk_config = config.config['disks'][disk_key]
                            backstore = disk_config['backstore']
                            backstore_object_name = disk_config['backstore_object_name']

                            lun = LUN(logger, pool, image_name,
                                      rbd_image.size(), local_gw, backstore,
                                      backstore_object_name)

                            if lun.error:
                                raise CephiSCSIError("Error defining rbd image {}"
                                                     .format(disk_key))

                            so = lun.allocate()
                            if lun.error:
                                raise CephiSCSIError("Unable to register {} "
                                                     "with LIO: {}"
                                                     .format(disk_key,
                                                             lun.error_msg))

                            # If not in use by another target on this gw
                            # clean up stale locks.
                            if so.status != 'activated':
                                RBDDev.rbd_lock_cleanup(logger, ips,
                                                        rbd_image)

                            target._map_lun(config, so, target_disks[disk_key])
                            if target.error:
                                raise CephiSCSIError("Mapping for {} failed: {}"
                                                     .format(disk_key,
                                                             target.error_msg))


def rados_pool(conf=None, pool=None):
    """
    determine if a given pool name is defined within the ceph cluster
    :param pool: pool name to check for (str)
    :return: Boolean representing the pool's existence
    """

    if conf is None:
        conf = settings.config.cephconf
    if pool is None:
        pool = settings.config.pool

    with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
        pool_list = cluster.list_pools()

    return pool in pool_list




__author__ = 'pcuzner@redhat.com'

try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

import hashlib
import json
import re

from ceph_iscsi_config.gateway_setting import (TGT_SETTINGS, SYS_SETTINGS,
                                               TCMU_SETTINGS)


# this module when imported preserves the global values
# defined by the init method allowing other classes to
# access common configuration settings
def init():
    global config
    config = Settings()


class Settings(object):
    _float_regex = re.compile(r"^[0-9]*\.{1}[0-9]$")
    _int_regex = re.compile(r"^[0-9]+$")

    @staticmethod
    def normalize_controls(raw_controls, settings_list):
        """
        Convert a controls dictionary from a json converted or a user input
        dictionary where the values are strings.
        """
        controls = {}

        for key, raw_value in raw_controls.items():
            setting = settings_list.get(key)
            if setting is None:
                raise ValueError("Supported controls: {}".format(",".join(settings_list.keys())))

            if raw_value in [None, '']:
                # Use the default/reset.
                controls[key] = None
                continue

            controls[key] = setting.normalize(raw_value)

        return controls

    exclude_from_hash = ["cluster_client_name",
                         "logger_level"
                         ]

    def __init__(self, conffile='/etc/ceph/iscsi-gateway.cfg'):

        self.error = False
        self.error_msg = ''

        config = ConfigParser()
        dataset = config.read(conffile)

        self._add_attrs_from_defs(SYS_SETTINGS)
        self._add_attrs_from_defs(TGT_SETTINGS)
        self._add_attrs_from_defs(TCMU_SETTINGS)

        if len(dataset) != 0:
            # If we have a file use it to override the defaults
            if config.has_section("config"):
                self._override_attrs_from_conf(config.items("config"),
                                               SYS_SETTINGS)

            if config.has_section("target"):
                all_settings = TGT_SETTINGS.copy()
                all_settings.update(TCMU_SETTINGS)

                self._override_attrs_from_conf(config.items("target"),
                                               all_settings)

        self.cephconf = '{}/{}.conf'.format(self.ceph_config_dir, self.cluster_name)
        if self.api_secure:
            self.api_ssl_verify = False if self.api_secure else None

    def __repr__(self):
        s = ''
        for k in self.__dict__:
            s += "{} = {}\n".format(k, self.__dict__[k])
        return s

    def _add_attrs_from_defs(self, def_settings):
        """
        receive a settings dict and apply those key/value to the
        current instance, settings that look like numbers are converted
        :param settings: array of setting objects
        :return: None
        """
        for k, setting in def_settings.items():
            self.__setattr__(k, setting.def_val)

    def _override_attrs_from_conf(self, config, def_settings):
        """
        receive a settings dict and apply those key/value to the
        current instance, settings that look like numbers are converted
        :param settings: dict of settings
        :return: None
        """
        for k, v in config:
            if hasattr(self, k):
                setting = def_settings[k]
                try:
                    self.__setattr__(k, setting.normalize(v))
                except ValueError:
                    # We do not even have the logger up yet, so just ignore
                    # so the deamons can still start
                    pass

    def _hash_settings(self, def_settings, sync_settings):
        for setting in def_settings:
            if setting not in Settings.exclude_from_hash:
                sync_settings[setting] = getattr(self, setting)

    def hash(self):
        """
        Generates a sha256 hash of the settings that are required to be in sync between gateways.
        :return: checksum (str)
        """

        sync_settings = {}
        self._hash_settings(SYS_SETTINGS.keys(), sync_settings)
        self._hash_settings(TGT_SETTINGS.keys(), sync_settings)
        self._hash_settings(TCMU_SETTINGS.keys(), sync_settings)

        h = hashlib.sha256()
        h.update(json.dumps(sync_settings).encode('utf-8'))
        return h.hexdigest()



import socket
import netifaces
import subprocess
import rados
import rbd
import re
import datetime
import os

import ceph_iscsi_config.settings as settings

__author__ = 'pcuzner@redhat.com'

size_suffixes = ['M', 'G', 'T']


class CephiSCSIError(Exception):
    '''
    Generic Ceph iSCSI config error.
    '''
    pass


class CephiSCSIInval(CephiSCSIError):
    '''
    Invalid setting/param.
    '''
    pass


def shellcommand(command_string):

    try:
        response = subprocess.check_output(command_string, shell=True)
    except subprocess.CalledProcessError:
        return None
    else:
        return response


def normalize_ip_address(ip_address):
    """
    IPv6 addresses should not include the square brackets utilized by
    IPv6 literals (RFC 3986)
    """
    address_regex = re.compile(r"^\[(.*)\]$")
    match = address_regex.match(ip_address)
    if match:
        return match.group(1)
    return ip_address


def normalize_ip_literal(ip_address):
    """
    rtslib expects IPv4 addresses as a dotted-quad string, and IPv6
    addresses surrounded by brackets.
    """
    ip_address = normalize_ip_address(ip_address)
    try:
        socket.inet_pton(socket.AF_INET6, ip_address)
        return "[" + ip_address + "]"
    except Exception:
        pass

    return ip_address


def resolve_ip_addresses(addr):
    """
    return list of IPv4/IPv6 address for the given address - could be an ip or
    name passed in
    :param addr: name or ip address (dotted quad)
    :return: list of IPv4/IPv6 addresses
    """
    families = [socket.AF_INET, socket.AF_INET6]
    normalized_addr = normalize_ip_address(addr)
    for family in families:
        try:
            socket.inet_pton(family, normalized_addr)
            return [normalized_addr]
        except Exception:
            pass

    addrs = set()
    try:
        infos = socket.getaddrinfo(addr, 0)
        for info in infos:
            if info[0] in families:
                addrs.add(info[4][0])
    except Exception:
        pass

    return list(addrs)


def valid_ip(ip, port=22):
    """
    Validate either a single IP or a list of IPs. An IP is valid if I can
    reach port 22 - since that's a common
    :param args:
    :return: Boolean
    """
    if isinstance(ip, str):
        ip_list = list([ip])
    elif isinstance(ip, list):
        ip_list = ip
    else:
        return False

    ip_ok = True

    families = [socket.AF_INET, socket.AF_INET6]
    for addr in ip_list:
        addr_ok = False
        for family in families:
            sock = socket.socket(family, socket.SOCK_STREAM)
            sock.settimeout(1)
            try:
                sock.connect((addr, port))
            except socket.error:
                pass
            else:
                sock.close()
                addr_ok = True
                break

        if not addr_ok:
            ip_ok = False
            break
    return ip_ok


def valid_size(size):
    valid = True
    unit = size[-1]

    if unit.upper() not in size_suffixes:
        valid = False
    else:
        try:
            int(size[:-1])
        except ValueError:
            valid = False

    return valid


def format_lio_yes_no(value):
    if value:
        return "Yes"
    return "No"


def ip_addresses():
    """
    return a list of IPv4/IPv6 addresses on the system (excluding 127.0.0.1/::1)
    :return: IP address list
    """
    ip_list = set()
    for iface in netifaces.interfaces():
        if netifaces.AF_INET in netifaces.ifaddresses(iface):
            for link in netifaces.ifaddresses(iface)[netifaces.AF_INET]:
                ip_list.add(link['addr'])
        if netifaces.AF_INET6 in netifaces.ifaddresses(iface):
            for link in netifaces.ifaddresses(iface)[netifaces.AF_INET6]:
                if '%' in link['addr']:
                    continue
                ip_list.add(link['addr'])

    ip_list.discard('::1')
    ip_list.discard('127.0.0.1')

    return list(ip_list)


def human_size(num):
    for unit, precision in [('b', 0), ('K', 0), ('M', 0), ('G', 0), ('T', 1),
                            ('P', 1), ('E', 2), ('Z', 2)]:
        if num % 1024 != 0:
            return "{0:.{1}f}{2}".format(num, precision, unit)
        num /= 1024.0
    return "{0:.2f}{1}".format(num, "Y")


def convert_2_bytes(disk_size):

    try:
        # If it's already an integer or a string with no suffix then assume
        # it's already in bytes.
        return int(disk_size)
    except ValueError:
        pass

    power = [2, 3, 4]
    unit = disk_size[-1].upper()
    offset = size_suffixes.index(unit)
    value = int(disk_size[:-1])  # already validated, so no need for try/except clause

    _bytes = value * (1024 ** power[offset])

    return _bytes


def get_pool_id(conf=None, pool_name=None):
    """
    Query Rados to get the pool id of a given pool name
    :param conf: ceph configuration file
    :param pool_name: pool name (str)
    :return: pool id (int)
    """

    if conf is None:
        conf = settings.config.cephconf
    if pool_name is None:
        pool_name = settings.config.pool

    with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
        pool_id = cluster.pool_lookup(pool_name)

    return pool_id


def get_pool_name(conf=None, pool_id=0):
    """
    Query Rados to get the pool name of a given pool_id
    :param conf: ceph configuration file
    :param pool_name: pool id number (int)
    :return: pool name (str)
    """

    if conf is None:
        conf = settings.config.cephconf

    with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
        pool_name = cluster.pool_reverse_lookup(pool_id)

    return pool_name


def get_rbd_size(pool, image, conf=None):
    """
    return the size of a given rbd from the local ceph cluster
    :param pool: (str) pool name
    :param image: (str) rbd image name
    :return: (int) size in bytes of the rbd
    """

    if conf is None:
        conf = settings.config.cephconf

    with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
        with cluster.open_ioctx(pool) as ioctx:
            with rbd.Image(ioctx, image) as rbd_image:
                size = rbd_image.size()
    return size


def get_pools(conf=None):
    """
    return a list of pools in the local ceph cluster
    :param conf: (str) or None
    :return: (list) of pool names
    """

    if conf is None:
        conf = settings.config.cephconf

    with rados.Rados(conffile=conf, name=settings.config.cluster_client_name) as cluster:
        pool_list = cluster.list_pools()

    return pool_list


def get_time():
    utc = datetime.datetime.utcnow()
    return utc.strftime('%Y/%m/%d %H:%M:%S')


def this_host():
    """
    return the local machine's fqdn
    """
    return socket.getfqdn()


def encryption_available():
    """
    Determine whether encryption is available by looking for the relevant
    keys
    :return: (bool) True if all keys are present, else False
    """
    encryption_keys = list([settings.config.priv_key,
                            settings.config.pub_key])

    config_dir = settings.config.ceph_config_dir
    keys = [os.path.join(config_dir, key_name)
            for key_name in encryption_keys]

    return all([os.path.exists(key) for key in keys])


def read_os_release():
    os_release_file = '/etc/os-release'
    d = {}
    if not os.path.exists(os_release_file):
        return d
    with open(os_release_file) as f:
        for line in f:
            rs = line.rstrip()
            if rs:
                k, v = rs.split("=")
                d[k] = v.strip('"')
    return d


def gen_control_string(controls):
    """
    Generate a kernel control string from a given dictionary
    of control arguments.
    :return: control string (str)
    """
    control = ''
    for key, value in controls.items():
        if value is not None:
            control += "{}={},".format(key, value)
    return None if control == '' else control[:-1]


class ListComparison(object):

    def __init__(self, current_list, new_list):
        """
        compare two lists to identify changes
        :param current_list : list of current values (existing state)
        :param new_list: list if new values (desired state)
        """
        self.current = current_list
        self.new = new_list
        self.changed = False

    @property
    def added(self):
        """
        provide a list of added items
        :return: (list) in the sequence provided
        """
        additions = set(self.new) - set(self.current)
        if len(additions) > 0:
            self.changed = True

        # simply returning the result of the set comparison does not preserve
        # the list item sequence. By iterating over the new list we can
        # return the expected sequence
        return [item for item in self.new if item in additions]

    @property
    def removed(self):
        """
        calculate the removed items between two lists using set comparisons
        :return: (list) removed items
        """
        removals = set(self.current) - set(self.new)
        if len(removals) > 0:
            self.changed = True
        return list(removals)



#!/usr/bin/python

from setuptools import setup

import distutils.command.install_scripts
import shutil
import os

if os.path.exists('README'):
    with open('README') as readme_file:
        long_description = readme_file.read().strip()
else:
    long_description = ''


# idea from http://stackoverflow.com/a/11400431/2139420
class StripExtension(distutils.command.install_scripts.install_scripts):
    """
    Class to handle the stripping of .py extensions in for executable file names
    making them more user friendly
    """
    def run(self):
        distutils.command.install_scripts.install_scripts.run(self)
        for script in self.get_outputs():
            if script.endswith(".py"):
                shutil.move(script, script[:-3])


setup(
    name="ceph_iscsi",
    version="3.4",
    description="Common classes/functions and CLI tools used to configure iSCSI "
                "gateways backed by Ceph RBD",
    long_description=long_description,
    author="Paul Cuzner",
    author_email="pcuzner@redhat.com",
    url="http://github.com/pcuzner/ceph-iscsi",
    license="GPLv3",
    packages=[
        "ceph_iscsi_config",
        "gwcli"
    ],
    scripts=[
        "rbd-target-gw.py",
        'gwcli.py',
        'rbd-target-api.py'
    ],
    data_files=[("/var/log/rbd-target-gw", []),
                ("/var/log/rbd-target-api", [])],
    cmdclass={
        "install_scripts": StripExtension
    }
)
